# encoding: utf-8
"""
operational.py

Created by Thomas Mangin on 2013-09-01.
Copyright (c) 2013-2015 Exa Networks. All rights reserved.
"""

from exabgp.bgp.message.open.capability.capability import Capability

# ================================================================== Operational
#


class Operational (Capability,list):
	ID = Capability.CODE.OPERATIONAL

	def __str__ (self):
		# XXX: FIXME: could be more verbose
		return 'Operational'

	def json (self):
		return '{ "name": "operational" }'

	def extract (self):
		return ['']

	@staticmethod
	def unpack_capability (instance, data, capability=None):  # pylint: disable=W0613
		# XXX: FIXME: we should set that that instance was seen and raise if seen twice
		return instance



# encoding: utf-8
"""
attributes.py

Created by Thomas Mangin on 2015-03-31.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from struct import pack
from struct import unpack
from struct import calcsize
from exabgp.netlink import NetLinkError


class AttributesError (NetLinkError):
	pass


class Attributes (object):
	class Header (object):
		PACK = 'HH'
		LEN = calcsize(PACK)

	class Type (object):
		IFA_UNSPEC     = 0x00
		IFA_ADDRESS    = 0x01
		IFA_LOCAL      = 0x02
		IFA_LABEL      = 0x03
		IFA_BROADCAST  = 0x04
		IFA_ANYCAST    = 0x05
		IFA_CACHEINFO  = 0x06
		IFA_MULTICAST  = 0x07

	@classmethod
	def decode (cls, data):
		while data:
			length, atype, = unpack(cls.Header.PACK,data[:cls.Header.LEN])
			if len(data) < length:
				raise AttributesError("Buffer underrun %d < %d" % (len(data),length))
			payload = data[cls.Header.LEN:length]
			yield atype, payload
			data = data[int((length + 3) / 4) * 4:]

	@classmethod
	def encode (cls, attributes):
		def _encode (atype, payload):
			def pad (length, to=4):
				return (length+to-1) & ~(to-1)
			length = cls.Header.LEN + len(payload)
			raw = pack(cls.Header.PACK,length,atype) + payload
			pad = pad(length) - len(raw)
			if pad:
				raw += '\0'*pad
			return raw

		return ''.join([_encode(k,v) for (k,v) in attributes.items()])



# encoding: utf-8
"""
community.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.bgp.message.update.attribute.community.extended.community import ExtendedCommunity

from exabgp.bgp.message.update.attribute.community.extended.encapsulation import Encapsulation
from exabgp.bgp.message.update.attribute.community.extended.l2info import L2Info
from exabgp.bgp.message.update.attribute.community.extended.origin import OriginASNIP
from exabgp.bgp.message.update.attribute.community.extended.origin import OriginIPASN
from exabgp.bgp.message.update.attribute.community.extended.origin import OriginASN4Number
from exabgp.bgp.message.update.attribute.community.extended.rt import RouteTargetASNIP
from exabgp.bgp.message.update.attribute.community.extended.rt import RouteTargetIPASN
from exabgp.bgp.message.update.attribute.community.extended.rt import RouteTargetASN4Number
from exabgp.bgp.message.update.attribute.community.extended.traffic import TrafficRate
from exabgp.bgp.message.update.attribute.community.extended.traffic import TrafficAction
from exabgp.bgp.message.update.attribute.community.extended.traffic import TrafficRedirect
from exabgp.bgp.message.update.attribute.community.extended.traffic import TrafficMark
from exabgp.bgp.message.update.attribute.community.extended.traffic import TrafficNextHop

ExtendedCommunity.register_extended(Encapsulation)
ExtendedCommunity.register_extended(L2Info)

ExtendedCommunity.register_extended(OriginASNIP)
ExtendedCommunity.register_extended(OriginIPASN)
ExtendedCommunity.register_extended(OriginASN4Number)

ExtendedCommunity.register_extended(RouteTargetASNIP)
ExtendedCommunity.register_extended(RouteTargetIPASN)
ExtendedCommunity.register_extended(RouteTargetASN4Number)

ExtendedCommunity.register_extended(TrafficRate)
ExtendedCommunity.register_extended(TrafficAction)
ExtendedCommunity.register_extended(TrafficRedirect)
ExtendedCommunity.register_extended(TrafficMark)
ExtendedCommunity.register_extended(TrafficNextHop)



# encoding: utf-8
"""
hashtable.py

Created by Thomas Mangin on 2009-09-06.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""


class HashTable (dict):
	def __getitem__ (self, key):
		return dict.__getitem__(self,key.replace('_','-'))

	def __setitem__ (self, key, value):
		return dict.__setitem__(self,key.replace('_','-'),value)

	def __getattr__ (self, key):
		return dict.__getitem__(self,key.replace('_','-'))

	def __setattr__ (self, key, value):
		return dict.__setitem__(self,key.replace('_','-'),value)



# encoding: utf-8
"""
usage.py

Created by Thomas Mangin on 2009-09-06.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

import sys
import resource

if sys.platform == 'darwin':
	# darwin returns bytes
	divisor = 1024.0*1024.0
else:
	# other OS (AFAIK) return a number of pages
	divisor = 1024.0*1024.0/resource.getpagesize()


def usage (label='usage'):
	rusage = resource.getrusage(resource.RUSAGE_SELF)
	return '%s: usertime=%s systime=%s mem=%s mb' % (label,rusage.ru_utime,rusage.ru_stime,(rusage.ru_maxrss/divisor))



# encoding: utf-8
"""
eor.py

Created by Thomas Mangin on 2010-01-16.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

# from struct import unpack

from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI
from exabgp.bgp.message import Message
from exabgp.bgp.message.update.attribute.attributes import Attributes
from exabgp.bgp.message.update.nlri.nlri import NLRI as _NLRI

# =================================================================== End-Of-RIB
# not technically a different message type but easier to treat as one


class EOR (Message):
	TYPE = chr(Message.CODE.UPDATE)

	class NLRI (_NLRI):
		PREFIX = '\x00\x00\x00\x07\x90\x0F\x00\x03'
		EOR = True

		nexthop = None

		def __init__ (self, afi, safi, action):
			_NLRI.__init__(self,afi,safi)
			self.action = action

		def pack (self):
			if self.afi == AFI.ipv4 and self.safi == SAFI.unicast:
				return '\x00\x00\x00\x00'
			return self.PREFIX + self.afi.pack() + self.safi.pack()

		def __str__ (self):
			return self.extensive()

		def extensive (self):
			return 'eor %d/%d (%s %s)' % (self.afi,self.safi,self.afi,self.safi)

		def json (self):
			return '"eor": { "afi" : "%s", "safi" : "%s" }' % (self.afi,self.safi)

	def __init__ (self, afi, safi, action=None):
		self.nlris = [EOR.NLRI(afi,safi,action),]
		self.attributes = Attributes()

	def message (self):
		return self._message(
			self.nlris[0].pack()
		)

	def __str__ (self):
		return 'EOR'

	@classmethod
	def unpack_message (cls, data, negotiated):
		header_length = len(EOR.NLRI.PREFIX)
		return cls(AFI.unpack(data[header_length:header_length+2]),SAFI.unpack(data[header_length+2]))



# encoding: utf-8
"""
open.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from struct import unpack

from exabgp.bgp.message import Message
from exabgp.bgp.message.open.version import Version
from exabgp.bgp.message.open.asn import ASN
from exabgp.bgp.message.open.holdtime import HoldTime
from exabgp.bgp.message.open.routerid import RouterID
from exabgp.bgp.message.open.capability.capabilities import Capabilities
from exabgp.bgp.message.notification import Notify

# =================================================================== Open

# 0                   1                   2                   3
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+
# |    Version    |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |     My Autonomous System      |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |           Hold Time           |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                         BGP Identifier                        |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# | Opt Parm Len  |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                                                               |
# |             Optional Parameters (variable)                    |
# |                                                               |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

# Optional Parameters:

# 0                   1
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-...
# |  Parm. Type   | Parm. Length  |  Parameter Value (variable)
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-...


class Open (Message):
	ID = Message.CODE.OPEN
	TYPE = chr(Message.CODE.OPEN)

	def __init__ (self, version, asn, router_id, capabilities, hold_time):
		self.version = Version(version)
		self.asn = ASN(asn)
		self.hold_time = HoldTime(hold_time)
		self.router_id = RouterID(router_id)
		self.capabilities = capabilities

	def message (self):
		return self._message("%s%s%s%s%s" % (
			self.version.pack(),
			self.asn.trans().pack(),
			self.hold_time.pack(),
			self.router_id.pack(),
			self.capabilities.pack()
		))

	def __str__ (self):
		return "OPEN version=%d asn=%d hold_time=%s router_id=%s capabilities=[%s]" % (self.version, self.asn.trans(), self.hold_time, self.router_id,self.capabilities)

	@classmethod
	def unpack_message (cls, data, _):
		version = ord(data[0])
		if version != 4:
			# Only version 4 is supported nowdays..
			raise Notify(2,1,data[0])
		asn = unpack('!H',data[1:3])[0]
		hold_time = unpack('!H',data[3:5])[0]
		numeric = unpack('!L',data[5:9])[0]
		router_id = "%d.%d.%d.%d" % (numeric >> 24,(numeric >> 16) & 0xFF,(numeric >> 8) & 0xFF,numeric & 0xFF)
		capabilities = Capabilities.unpack(data[9:])
		return cls(version,asn,router_id,capabilities,hold_time)



#!/usr/bin/python

# based on netlink.py at ....
# https://gforge.inria.fr/scm/viewvc.php/canso/trunk/tools/netlink.py?view=markup&revision=1360&root=mehani&pathrev=1360
# http://www.linuxjournal.com/article/7356?page=0,1
# http://smacked.org/docs/netlink.pdf
# RFC 3549

import os
import socket
from struct import pack
from struct import unpack
from struct import calcsize
from collections import namedtuple

try:
	getattr(socket,'AF_NETLINK')
except AttributeError:
	raise ImportError('This module only works on unix version with netlink support')


class NetMask (object):
	CIDR = {
		32: '255.255.255.255',
		31: '255.255.255.254',
		30: '255.255.255.252',
		29: '255.255.255.248',
		28: '255.255.255.240',
		27: '255.255.255.224',
		26: '255.255.255.192',
		25: '255.255.255.128',
		24: '255.255.255.0',
		23: '255.255.254.0',
		22: '255.255.252.0',
		21: '255.255.248.0',
		20: '255.255.240.0',
		19: '255.255.224.0',
		18: '255.255.192.0',
		17: '255.255.128.0',
		16: '255.255.0.0',
		15: '255.254.0.0',
		14: '255.252.0.0',
		13: '255.248.0.0',
		12: '255.240.0.0',
		11: '255.224.0.0',
		10: '255.192.0.0',
		9: '255.128.0.0',
		8: '255.0.0.0',
		7: '254.0.0.0',
		6: '252.0.0.0',
		5: '248.0.0.0',
		4: '240.0.0.0',
		3: '224.0.0.0',
		2: '192.0.0.0',
		1: '128.0.0.0',
		0: '0.0.0.0',
	}


class GlobalError (Exception):
	pass


class NetLinkError (GlobalError):
	pass


class Sequence (int):
	_instance = dict()

	def __new__ (cls):
		cls._instance['next'] = cls._instance.get('next', 0) + 1
		return cls._instance['next']


class NetLinkRoute (object):
	_IGNORE_SEQ_FAULTS = True

	NETLINK_ROUTE = 0

	format = namedtuple('Message','type flags seq pid data')
	pid = os.getpid()
	netlink = socket.socket(socket.AF_NETLINK, socket.SOCK_RAW, NETLINK_ROUTE)

	class Header (object):
		# linux/netlink.h
		PACK  = 'IHHII'
		LEN = calcsize(PACK)

	class Command (object):
		NLMSG_NOOP    = 0x01
		NLMSG_ERROR   = 0x02
		NLMSG_DONE    = 0x03
		NLMSG_OVERRUN = 0x04

	class Flags (object):
		NLM_F_REQUEST = 0x01  # It is query message.
		NLM_F_MULTI   = 0x02  # Multipart message, terminated by NLMSG_DONE
		NLM_F_ACK     = 0x04  # Reply with ack, with zero or error code
		NLM_F_ECHO    = 0x08  # Echo this query

		# Modifiers to GET query
		NLM_F_ROOT   = 0x100  # specify tree root
		NLM_F_MATCH  = 0x200  # return all matching
		NLM_F_DUMP   = NLM_F_ROOT | NLM_F_MATCH
		NLM_F_ATOMIC = 0x400  # atomic GET

		# Modifiers to NEW query
		NLM_F_REPLACE = 0x100  # Override existing
		NLM_F_EXCL    = 0x200  # Do not touch, if it exists
		NLM_F_CREATE  = 0x400  # Create, if it does not exist
		NLM_F_APPEND  = 0x800  # Add to end of list

	errors = {
		Command.NLMSG_ERROR:   'netlink error',
		Command.NLMSG_OVERRUN: 'netlink overrun',
	}

	@classmethod
	def encode (cls, dtype, seq, flags, body, attributes):
		attrs = Attributes.encode(attributes)
		length = cls.Header.LEN + len(attrs) + len(body)
		return pack(cls.Header.PACK, length, dtype, flags, seq, cls.pid) + body + attrs

	@classmethod
	def decode (cls, data):
		while data:
			length, ntype, flags, seq, pid = unpack(cls.Header.PACK,data[:cls.Header.LEN])
			if len(data) < length:
				raise NetLinkError("Buffer underrun")
			yield cls.format(ntype, flags, seq, pid, data[cls.Header.LEN:length])
			data = data[length:]

	@classmethod
	def send (cls, dtype, hflags, family=socket.AF_UNSPEC):
		sequence = Sequence()

		message = cls.encode(
			dtype,
			sequence,
			hflags,
			pack('Bxxx', family),
			{}
		)

		cls.netlink.send(message)

		while True:
			data = cls.netlink.recv(640000)
			for mtype, flags, seq, pid, data in cls.decode(data):
				if seq != sequence:
					if cls._IGNORE_SEQ_FAULTS:
						continue
					raise NetLinkError("netlink seq mismatch")
				if mtype == NetLinkRoute.Command.NLMSG_DONE:
					raise StopIteration()
				elif dtype in cls.errors:
					raise NetLinkError(cls.errors[mtype])
				else:
					yield data

	# def change (self, dtype, family=socket.AF_UNSPEC):
	# 	for _ in self.send(dtype, self.Flags.NLM_F_REQUEST | self.Flags.NLM_F_CREATE,family):
	# 		yield _


class AttributesError (GlobalError):
	pass


class Attributes (object):
	class Header (object):
		PACK = 'HH'
		LEN = calcsize(PACK)

	class Type (object):
		IFA_UNSPEC     = 0x00
		IFA_ADDRESS    = 0x01
		IFA_LOCAL      = 0x02
		IFA_LABEL      = 0x03
		IFA_BROADCAST  = 0x04
		IFA_ANYCAST    = 0x05
		IFA_CACHEINFO  = 0x06
		IFA_MULTICAST  = 0x07

	@classmethod
	def decode (cls, data):
		while data:
			length, atype, = unpack(cls.Header.PACK,data[:cls.Header.LEN])
			if len(data) < length:
				raise AttributesError("Buffer underrun %d < %d" % (len(data),length))
			payload = data[cls.Header.LEN:length]
			yield atype, payload
			data = data[int((length + 3) / 4) * 4:]

	@classmethod
	def encode (cls, attributes):
		def _encode (atype, payload):
			def pad (length, to=4):
				return (length+to-1) & ~(to-1)
			length = cls.Header.LEN + len(payload)
			raw = pack(cls.Header.PACK,length,atype) + payload
			pad = pad(length) - len(raw)
			if pad:
				raw += '\0'*pad
			return raw

		return ''.join([_encode(k,v) for (k,v) in attributes.items()])


class _InfoMessage (object):
	# to be defined by the subclasses
	format = namedtuple('Parent', 'to be subclassed')

	# to be defined by the subclasses
	class Header (object):
		PACK = ''
		LEN = 0

	def __init__ (self, route):
		self.route = route

	def decode (self, data):
		extracted = list(unpack(self.Header.PACK,data[:self.Header.LEN]))
		attributes = Attributes.decode(data[self.Header.LEN:])
		extracted.append(dict(attributes))
		return self.format(*extracted)

	def extract (self, atype, flags=NetLinkRoute.Flags.NLM_F_REQUEST | NetLinkRoute.Flags.NLM_F_DUMP, family=socket.AF_UNSPEC):
		for data in self.route.send(atype,flags,family):
			yield self.decode(data)


# 0                   1                   2                   3
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |   Family    |   Reserved  |          Device Type              |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                     Interface Index                           |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                      Device Flags                             |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                      Change Mask                              |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class Link(_InfoMessage):
	class Header (object):
		PACK = 'BxHiII'
		LEN = calcsize(PACK)

	# linux/if_link.h
	format = namedtuple('Info', 'family type index flags change attributes')

	class Command (object):
		# linux/rtnetlink.h
		RTM_NEWLINK = 0x10  # Create a new network interface
		RTM_DELLINK = 0x11  # Destroy a network interface
		RTM_GETLINK = 0x12  # Retrieve information about a network interface (ifinfomsg)
		RTM_SETLINK = 0x13  # -

	class Type (object):
		class Family (object):
			AF_INET  = socket.AF_INET
			AF_INET6 = socket.AF_INET6

		class Device (object):
			IFF_UP            = 0x0001  # Interface is administratively up.
			IFF_BROADCAST     = 0x0002  # Valid broadcast address set.
			IFF_DEBUG         = 0x0004  # Internal debugging flag.
			IFF_LOOPBACK      = 0x0008  # Interface is a loopback interface.
			IFF_POINTOPOINT   = 0x0010  # Interface is a point-to-point link.
			IFF_NOTRAILERS    = 0x0020  # Avoid use of trailers.
			IFF_RUNNING       = 0x0040  # Interface is operationally up.
			IFF_NOARP         = 0x0080  # No ARP protocol needed for this interface.
			IFF_PROMISC       = 0x0100  # Interface is in promiscuous mode.
			IFF_ALLMULTI      = 0x0200  # Receive all multicast packets.
			IFF_MASTER        = 0x0400  # Master of a load balancing bundle.
			IFF_SLAVE         = 0x0800  # Slave of a load balancing bundle.
			IFF_MULTICAST     = 0x1000  # Supports multicast.

			IFF_PORTSEL       = 0x2000  # Is able to select media type via ifmap.
			IFF_AUTOMEDIA     = 0x4000  # Auto media selection active.
			IFF_DYNAMIC       = 0x8000  # Interface was dynamically created.

			IFF_LOWER_UP      = 0x10000  # driver signals L1 up
			IFF_DORMANT       = 0x20000  # driver signals dormant
			IFF_ECHO          = 0x40000  # echo sent packet

		class Attribute (object):
			IFLA_UNSPEC      = 0x00
			IFLA_ADDRESS     = 0x01
			IFLA_BROADCAST   = 0x02
			IFLA_IFNAME      = 0x03
			IFLA_MTU         = 0x04
			IFLA_LINK        = 0x05
			IFLA_QDISC       = 0x06
			IFLA_STATS       = 0x07

	def getLinks (self):
		return self.extract(Link.Command.RTM_GETLINK)


# 0                   1                   2                   3
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |   Family    |     Length    |     Flags     |    Scope      |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                     Interface Index                         |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class Address (_InfoMessage):
	class Header (object):
		PACK = '4Bi'
		LEN = calcsize(PACK)

	format = namedtuple('Address', 'family prefixlen flags scope index attributes')

	class Command (object):
		RTM_NEWADDR = 0x14
		RTM_DELADDR = 0x15
		RTM_GETADDR = 0x16

	class Type (object):
		class Family (object):
			AF_INET  = socket.AF_INET
			AF_INET6 = socket.AF_INET6

		class Flag (object):
			IFA_F_SECONDARY  = 0x00  # For secondary address (alias interface)
			IFA_F_PERMANENT  = 0x00  # For a permanent address set by the user.  When this is not set, it means the address was dynamically created (e.g., by stateless autoconfiguration).
			IFA_F_DEPRECATED = 0x00  # Defines deprecated (IPV4) address
			IFA_F_TENTATIVE  = 0x00  # Defines tentative (IPV4) address (duplicate address detection is still in progress)

		class Scope (object):
			RT_SCOPE_UNIVERSE = 0x00  # Global route
			RT_SCOPE_SITE     = 0x00  # Interior route in the local autonomous system
			RT_SCOPE_LINK     = 0x00  # Route on this link
			RT_SCOPE_HOST     = 0x00  # Route on the local host
			RT_SCOPE_NOWHERE  = 0x00  # Destination does not exist

		class Attribute (object):
			IFLA_UNSPEC      = 0x00
			IFLA_ADDRESS     = 0x01
			IFLA_BROADCAST   = 0x02
			IFLA_IFNAME      = 0x03
			IFLA_MTU         = 0x04
			IFLA_LINK        = 0x05
			IFLA_QDISC       = 0x06
			IFLA_STATS       = 0x07
			IFLA_COST        = 0x08
			IFLA_PRIORITY    = 0x09
			IFLA_MASTER      = 0x0A
			IFLA_WIRELESS    = 0x0B
			IFLA_PROTINFO    = 0x0C
			IFLA_TXQLEN      = 0x0D
			IFLA_MAP         = 0x0E
			IFLA_WEIGHT      = 0x0F
			IFLA_OPERSTATE   = 0x10
			IFLA_LINKMODE    = 0x11
			IFLA_LINKINFO    = 0x12
			IFLA_NET_NS_PID  = 0x13
			IFLA_IFALIAS     = 0x14
			IFLA_NUM_VF      = 0x15
			IFLA_VFINFO_LIST = 0x16
			IFLA_STATS64     = 0x17
			IFLA_VF_PORTS    = 0x18
			IFLA_PORT_SELF   = 0x19

	def getAddresses (self):
		return self.extract(Address.Command.RTM_GETADDR)


# 0                   1                   2                   3
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |   Family    |    Reserved1  |           Reserved2           |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                     Interface Index                         |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |           State             |     Flags     |     Type      |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class Neighbor (_InfoMessage):
	class Header (object):
		# linux/if_addr.h
		PACK = 'BxxxiHBB'
		LEN = calcsize(PACK)

	format = namedtuple('Neighbor', 'family index state flags type attributes')

	class Command (object):
		RTM_NEWNEIGH = 0x1C
		RTM_DELNEIGH = 0x1D
		RTM_GETNEIGH = 0x1E

	class Type (object):
		class Family (object):
			AF_INET  = socket.AF_INET
			AF_INET6 = socket.AF_INET6

		class State (object):
			NUD_INCOMPLETE = 0x01  # Still attempting to resolve
			NUD_REACHABLE  = 0x02  # A confirmed working cache entry
			NUD_STALE      = 0x04  # an expired cache entry
			NUD_DELAY      = 0x08  # Neighbor no longer reachable.  Traffic sent, waiting for confirmatio.
			NUD_PROBE      = 0x10  # A cache entry that is currently being re-solicited
			NUD_FAILED     = 0x20  # An invalid cache entry
			# Dummy states
			NUD_NOARP      = 0x40  # A device which does not do neighbor discovery (ARP)
			NUD_PERMANENT  = 0x80  # A static entry
			NUD_NONE       = 0x00

		class Flag (object):
			NTF_USE        = 0x01
			NTF_PROXY      = 0x08  # A proxy ARP entry
			NTF_ROUTER     = 0x80  # An IPv6 router

		class Attribute (object):
			# XXX : Not sure - starts at zero or one ... ??
			NDA_UNSPEC     = 0x00  # Unknown type
			NDA_DST        = 0x01  # A neighbour cache network. layer destination address
			NDA_LLADDR     = 0x02  # A neighbor cache link layer address.
			NDA_CACHEINFO  = 0x03  # Cache statistics
			NDA_PROBES     = 0x04

	def getNeighbors (self):
		return self.extract(Neighbor.Command.RTM_GETNEIGH)


# 0                   1                   2                   3
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |   Family    |  Src length   |  Dest length  |     TOS       |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |  Table ID   |   Protocol    |     Scope     |     Type      |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                          Flags                              |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class Network (_InfoMessage):
	class Header (object):
		# linux/if_addr.h
		PACK = '8BI'  # or is it 8Bi ?
		LEN = calcsize(PACK)

	format = namedtuple('Neighbor', 'family src_len dst_len tos table proto scope type flags attributes')

	class Command (object):
		RTM_NEWROUTE = 0x18
		RTM_DELROUTE = 0x19
		RTM_GETROUTE = 0x1A

	class Type (object):
		class Table (object):
			RT_TABLE_UNSPEC   = 0x00  # An unspecified routing table
			RT_TABLE_DEFAULT  = 0xFD  # The default table
			RT_TABLE_MAIN     = 0xFE  # The main table
			RT_TABLE_LOCAL    = 0xFF  # The local table

		class Protocol (object):
			RTPROT_UNSPEC     = 0x00  # Identifies what/who added the route
			RTPROT_REDIRECT   = 0x01  # By an ICMP redirect
			RTPROT_KERNEL     = 0x02  # By the kernel
			RTPROT_BOOT       = 0x03  # During bootup
			RTPROT_STATIC     = 0x04  # By the administrator
			RTPROT_GATED      = 0x08  # GateD
			RTPROT_RA         = 0x09  # RDISC/ND router advertissements
			RTPROT_MRT        = 0x0A  # Merit MRT
			RTPROT_ZEBRA      = 0x0B  # ZEBRA
			RTPROT_BIRD       = 0x0C  # BIRD
			RTPROT_DNROUTED   = 0x0D  # DECnet routing daemon
			RTPROT_XORP       = 0x0E  # XORP
			RTPROT_NTK        = 0x0F  # Netsukuku
			RTPROT_DHCP       = 0x10  # DHCP client
			# Self allocating ourself a kernel table :p
			# http://lxr.free-electrons.com/source/include/uapi/linux/rtnetlink.h#L237
			# YES WE CAN !
			RTPROT_EXABGP     = 0x11  # Exa Networks ExaBGP

		class Scope (object):
			RT_SCOPE_UNIVERSE = 0x00  # Global route
			RT_SCOPE_SITE     = 0xC8  # Interior route in the local autonomous system
			RT_SCOPE_LINK     = 0xFD  # Route on this link
			RT_SCOPE_HOST     = 0xFE  # Route on the local host
			RT_SCOPE_NOWHERE  = 0xFF  # Destination does not exist

		class Type (object):
			RTN_UNSPEC        = 0x00  # Unknown route.
			RTN_UNICAST       = 0x01  # A gateway or direct route.
			RTN_LOCAL         = 0x02  # A local interface route.
			RTN_BROADCAST     = 0x03  # A local broadcast route (sent as a broadcast).
			RTN_ANYCAST       = 0x04  # An anycast route.
			RTN_MULTICAST     = 0x05  # A multicast route.
			RTN_BLACKHOLE     = 0x06  # A silent packet dropping route.
			RTN_UNREACHABLE   = 0x07  # An unreachable destination.  Packets dropped and host unreachable ICMPs are sent to the originator.
			RTN_PROHIBIT      = 0x08  # A packet rejection route.  Packets are dropped and communication prohibited ICMPs are sent to the originator.
			RTN_THROW         = 0x09  # When used with policy routing, continue routing lookup in another table.  Under normal routing, packets are dropped and net unreachable ICMPs are sent to the originator.
			RTN_NAT           = 0x0A  # A network address translation rule.
			RTN_XRESOLVE      = 0x0B  # Refer to an external resolver (not implemented).

		class Flag (object):
			RTM_F_NOTIFY      = 0x100  # If the route changes, notify the user
			RTM_F_CLONED      = 0x200  # Route is cloned from another route
			RTM_F_EQUALIZE    = 0x400  # Allow randomization of next hop path in multi-path routing (currently not implemented)
			RTM_F_PREFIX      = 0x800  # Prefix Address

		class Attribute (object):
			RTA_UNSPEC        = 0x00  # Ignored.
			RTA_DST           = 0x01  # Protocol address for route destination address.
			RTA_SRC           = 0x02  # Protocol address for route source address.
			RTA_IIF           = 0x03  # Input interface index.
			RTA_OIF           = 0x04  # Output interface index.
			RTA_GATEWAY       = 0x05  # Protocol address for the gateway of the route
			RTA_PRIORITY      = 0x06  # Priority of route.
			RTA_PREFSRC       = 0x07  # Preferred source address in cases where more than one source address could be used.
			RTA_METRICS       = 0x08  # Route metrics attributed to route and associated protocols (e.g., RTT, initial TCP window, etc.).
			RTA_MULTIPATH     = 0x09  # Multipath route next hop's attributes.
			# RTA_PROTOINFO   = 0x0A  # Firewall based policy routing attribute.
			RTA_FLOW          = 0x0B  # Route realm.
			RTA_CACHEINFO     = 0x0C  # Cached route information.
			# RTA_SESSION     = 0x0D
			# RTA_MP_ALGO     = 0x0E
			RTA_TABLE         = 0x0F

	def getRoutes (self):
		return self.extract(Network.Command.RTM_GETROUTE)

	def _create (self, family):
		for _ in self.extract(Network.Command.RTM_NEWROUTE, flags,family):
			yield _

	def newRoute (self):
		network_flags  = NetLinkRoute.Flags.NLM_F_REQUEST
		network_flags |= NetLinkRoute.Flags.NLM_F_CREATE
		network_flags |= NetLinkRoute.Flags.NLM_F_EXCL
		# network_flags |= NetLinkRoute.Flags.NLM_F_ACK

		family = socket.AF_INET

		attributes = Attributes()
		Network.Attribute.RTA_DST

		neighbor = self.format(
			family,
			0,     # src_len
			32,    # dst_len ( only /32 atm)
			0,     # tos
			Network.Type.Table.RT_TABLE_MAIN,
			Network.Type.Protocol.RTPROT_EXABGP,
			Network.Type.Scope.RT_SCOPE_UNIVERSE,
			0,     # type
			Network.Flag.RTM_F_PREFIX,  # this may be wrong
			attributes
		)


		# prefix = '\20\x7f\0\0\2'


		for _ in self.extract(Network.Command.RTM_NEWROUTE,network_flags,family):
			yield _

	def delRoute (self):
		return self.extract(Network.Command.RTM_DELROUTE)

# 0                   1                   2                   3
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |   Family    |  Reserved1    |         Reserved2             |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                     Interface Index                         |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                      Qdisc handle                           |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                     Parent Qdisc                            |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                        TCM Info                             |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+


class TC (_InfoMessage):
	class Header (object):
		PACK = "BxxxiIII"
		LEN = calcsize(PACK)

	class Command (object):
		RTM_NEWQDISC = 36
		RTM_DELQDISC = 37
		RTM_GETQDISC = 38

	class Type (object):
		class Attribute (object):
			TCA_UNSPEC  = 0x00
			TCA_KIND    = 0x01
			TCA_OPTIONS = 0x02
			TCA_STATS   = 0x03
			TCA_XSTATS  = 0x04
			TCA_RATE    = 0x05
			TCA_FCNT    = 0x06
			TCA_STATS2  = 0x07


# 0                   1                   2                   3
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |   Mode    |    Reserved1  |           Reserved2             |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                         Range                               |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+


# 0                   1                   2                   3
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                       Packet ID                             |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                          Mark                               |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                       timestamp_m                           |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                       timestamp_u                           |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                          hook                               |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                       indev_name                            |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                       outdev_name                           |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |           hw_protocol       |        hw_type                |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |         hw_addrlen          |           Reserved            |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                       hw_addr                               |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                       data_len                              |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                      Payload . . .                          |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+


class Firewall (_InfoMessage):
	class Header (object):
		PACK = "BxxxI"
		LEN = calcsize(PACK)

	class Packet (object):
		class Header (object):
			PACK = "IIIIIIIHHHHII"
			LEN = calcsize(PACK)



# encoding: utf-8
"""
update/__init__.py

Created by Thomas Mangin on 2010-01-15.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from struct import pack

# ================================================================== BGP Message
#


# 0                   1                   2                   3
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                                                               |
# +                                                               +
# |                                                               |
# +                                                               +
# |                           Marker                              |
# +                                                               +
# |                                                               |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |          Length               |      Type     |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class Message (Exception):
	# we need to define TYPE inside __init__ of the subclasses
	# otherwise we can not dynamically create different UnknownMessage
	# TYPE = None

	MARKER = chr(0xff)*16
	HEADER_LEN = 19
	MAX_LEN = 4096

	registered_message = {}
	# This is redefined by the Notify class, Exception is never used
	klass_notify = Exception
	klass_unknown = Exception

	class CODE (int):
		__slots__ = []

		NOP           = 0x00  # .   0 - internal
		OPEN          = 0x01  # .   1
		UPDATE        = 0x02  # .   2
		NOTIFICATION  = 0x03  # .   3
		KEEPALIVE     = 0x04  # .   4
		ROUTE_REFRESH = 0x05  # .   5
		OPERATIONAL   = 0x06  # .   6  # Not IANA assigned yet

		names = {
			NOP:            'NOP',
			OPEN:           'OPEN',
			UPDATE:         'UPDATE',
			NOTIFICATION:   'NOTIFICATION',
			KEEPALIVE:      'KEEPALIVE',
			ROUTE_REFRESH:  'ROUTE_REFRESH',
			OPERATIONAL:    'OPERATIONAL',
		}

		def __str__ (self):
			return self.names.get(self,'unknown message %s' % hex(self))

		def __repr__ (self):
			return str(self)

		@staticmethod
		def name (message_id):
			return Message.CODE.names.get(message_id,'unknown message %s' % hex(message_id))

	Length = {
		CODE.OPEN:           lambda _:  _ >= 29,  # noqa
		CODE.UPDATE:         lambda _:  _ >= 23,  # noqa
		CODE.NOTIFICATION:   lambda _:  _ >= 21,  # noqa
		CODE.KEEPALIVE:      lambda _:  _ == 19,  # noqa
		CODE.ROUTE_REFRESH:  lambda _:  _ == 23,  # noqa
	}

	def __init__ (self):
		self._name = None

	@staticmethod
	def string (code):
		if code is None:
			return 'invalid'
		if code == Message.CODE.OPEN:
			return 'open'
		if code == Message.CODE.UPDATE:
			return 'update'
		if code == Message.CODE.NOTIFICATION:
			return 'notification'
		if code == Message.CODE.KEEPALIVE:
			return 'keepalive'
		if code == Message.CODE.ROUTE_REFRESH:
			return 'route refresh'
		if code == Message.CODE.OPERATIONAL:
			return 'operational'
		return 'unknown'

	def _message (self, message):
		message_len = pack('!H',19+len(message))
		return "%s%s%s%s" % (self.MARKER,message_len,self.TYPE,message)

	def message (self):
		raise NotImplementedError('message not implemented in subclasses')

	@staticmethod
	def register_message (klass, message=None):
		what = klass.TYPE if message is None else message
		if what in Message.registered_message:
			raise RuntimeError('only one class can be registered per message')
		Message.registered_message[ord(what)] = klass

	@classmethod
	def klass (cls, what):
		if what in cls.registered_message:
			return cls.registered_message[what]
		raise cls.klass_notify(2,4,'can not handle message %s' % what)

	@classmethod
	def unpack (cls, message, data, negotiated):
		if message in cls.registered_message:
			return cls.klass(message).unpack_message(data,negotiated)
		return cls.klass_unknown(message,data,negotiated)



# encoding: utf-8
"""
mprnlri.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from struct import unpack

from exabgp.protocol.ip import NoIP
from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI
from exabgp.protocol.ip.address import Address

from exabgp.bgp.message import IN
from exabgp.bgp.message.update.attribute.attribute import Attribute
from exabgp.bgp.message.update.nlri.nlri import NLRI

from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.open.capability.negotiated import Negotiated


# ==================================================== MP Unreacheable NLRI (15)
#

class MPRNLRI (Attribute,Address):
	FLAG = Attribute.Flag.OPTIONAL
	ID = Attribute.CODE.MP_REACH_NLRI

	# __slots__ = ['nlris']

	def __init__ (self, afi, safi, nlris):
		Address.__init__(self,afi,safi)
		# all the routes must have the same next-hop
		self.nlris = nlris

	def packed_attributes (self, addpath, maximum):
		if not self.nlris:
			return

		mpnlri = {}
		for nlri in self.nlris:
			if nlri.nexthop is NoIP:
				# EOR and Flow may not have any next_hop
				nexthop = ''
			else:
				# we do not want a next_hop attribute packed (with the _attribute()) but just the next_hop itself
				if nlri.safi.has_rd():
					# .packed and not .pack()
					nexthop = chr(0)*8 + nlri.nexthop.packed
				else:
					# .packed and not .pack()
					nexthop = nlri.nexthop.packed

			# mpunli[afi,safi][nexthop] = nlri
			mpnlri.setdefault((nlri.afi.pack(),nlri.safi.pack()),{}).setdefault(nexthop,[]).append(nlri.pack(addpath))

		for (pafi,psafi),data in mpnlri.iteritems():
			for nexthop,nlris in data.iteritems():
				payload = \
					pafi + psafi + \
					chr(len(nexthop)) + nexthop + \
					chr(0) + ''.join(nlris)

				if self._len(payload) <= maximum:
					yield self._attribute(payload)
					continue

				# This will not generate an optimum update size..
				# we should feedback the maximum on each iteration

				for nlri in nlris:
					yield self._attribute(
						pafi + psafi +
						chr(len(nexthop)) + nexthop +
						chr(0) + nlri
					)

	def pack (self, addpath):
		return ''.join(self.packed_attributes(addpath,Negotiated.FREE_SIZE))

	def __len__ (self):
		raise RuntimeError('we can not give you the size of an MPRNLRI - was it with our witout addpath ?')
		# return len(self.pack(False))

	def __str__ (self):
		return "MP_REACH_NLRI for %s %s with %d NLRI(s)" % (self.afi,self.safi,len(self.nlris))

	@classmethod
	def unpack (cls, data, negotiated):
		nlris = []

		# -- Reading AFI/SAFI
		afi,safi = unpack('!HB',data[:3])
		offset = 3

		# we do not want to accept unknown families
		if negotiated and (afi,safi) not in negotiated.families:
			raise Notify(3,0,'presented a non-negotiated family %d/%d' % (afi,safi))

		# -- Reading length of next-hop
		len_nh = ord(data[offset])
		offset += 1

		rd = 0

		# check next-hope size
		if afi == AFI.ipv4:
			if safi in (SAFI.unicast,SAFI.multicast):
				if len_nh != 4:
					raise Notify(3,0,'invalid ipv4 unicast/multicast next-hop length %d expected 4' % len_nh)
			elif safi in (SAFI.mpls_vpn,):
				if len_nh != 12:
					raise Notify(3,0,'invalid ipv4 mpls_vpn next-hop length %d expected 12' % len_nh)
				rd = 8
			elif safi in (SAFI.flow_ip,):
				if len_nh not in (0,4):
					raise Notify(3,0,'invalid ipv4 flow_ip next-hop length %d expected 4' % len_nh)
			elif safi in (SAFI.flow_vpn,):
				if len_nh not in (0,4):
					raise Notify(3,0,'invalid ipv4 flow_vpn next-hop length %d expected 4' % len_nh)
		elif afi == AFI.ipv6:
			if safi in (SAFI.unicast,):
				if len_nh not in (16,32):
					raise Notify(3,0,'invalid ipv6 unicast next-hop length %d expected 16 or 32' % len_nh)
			elif safi in (SAFI.mpls_vpn,):
				if len_nh not in (24,40):
					raise Notify(3,0,'invalid ipv6 mpls_vpn next-hop length %d expected 24 or 40' % len_nh)
				rd = 8
			elif safi in (SAFI.flow_ip,):
				if len_nh not in (0,16,32):
					raise Notify(3,0,'invalid ipv6 flow_ip next-hop length %d expected 0, 16 or 32' % len_nh)
			elif safi in (SAFI.flow_vpn,):
				if len_nh not in (0,16,32):
					raise Notify(3,0,'invalid ipv6 flow_vpn next-hop length %d expected 0, 16 or 32' % len_nh)
		size = len_nh - rd

		# XXX: FIXME: GET IT FROM CACHE HERE ?
		nhs = data[offset+rd:offset+rd+size]
		nexthops = [nhs[pos:pos+16] for pos in range(0,len(nhs),16)]

		# chech the RD is well zero
		if rd and sum([int(ord(_)) for _ in data[offset:8]]) != 0:
			raise Notify(3,0,"MP_REACH_NLRI next-hop's route-distinguisher must be zero")

		offset += len_nh

		# Skip a reserved bit as somone had to bug us !
		reserved = ord(data[offset])
		offset += 1

		if reserved != 0:
			raise Notify(3,0,'the reserved bit of MP_REACH_NLRI is not zero')

		# Is the peer going to send us some Path Information with the route (AddPath)
		addpath = negotiated.addpath.receive(afi,safi)

		# Reading the NLRIs
		data = data[offset:]

		if not data:
			raise Notify(3,0,'No data to decode in an MPREACHNLRI but it is not an EOR %d/%d' % (afi,safi))

		while data:
			if nexthops:
				for nexthop in nexthops:
					length,nlri = NLRI.unpack(afi,safi,data,addpath,nexthop,IN.ANNOUNCED)
					nlris.append(nlri)
			else:
				length,nlri = NLRI.unpack(afi,safi,data,addpath,'',IN.ANNOUNCED)
				nlris.append(nlri)

			# logger.parser(LazyFormat("parsed announce mp nlri %s payload " % nlri,data[:length]))
			data = data[length:]
		return cls(afi,safi,nlris)

EMPTY_MPRNLRI  = MPRNLRI(AFI(AFI.undefined),SAFI(SAFI.undefined),[])



# encoding: utf-8
"""
protocol.py

Created by Thomas Mangin on 2010-01-15.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.protocol.enum import Enum

# ===================================================================== Protocol
# http://www.iana.org/assignments/protocol-numbers/

class Protocol (Enum):
	ICMP  = 0x01
	IGMP  = 0x02
	TCP   = 0x06
	EGP   = 0x08
	UDP   = 0x11
	RSVP  = 0x2E
	GRE   = 0x2F
	ESP   = 0x32
	AH    = 0x33
	OSPF  = 0x59
	IPIP  = 0x5E
	PIM   = 0x67
	SCTP  = 0x84

	def pack (self):
		return chr(self)


Protocol.UNKNOWN = "unknown protocol %s"

Protocol.NAME = {
	Protocol(Protocol.ICMP):  'icmp',
	Protocol(Protocol.IGMP):  'igmp',
	Protocol(Protocol.TCP):   'tcp',
	Protocol(Protocol.EGP):   'egp',
	Protocol(Protocol.UDP):   'udp',
	Protocol(Protocol.RSVP):  'rsvp',
	Protocol(Protocol.GRE):   'gre',
	Protocol(Protocol.ESP):   'esp',
	Protocol(Protocol.AH):    'ah',
	Protocol(Protocol.OSPF):  'ospf',
	Protocol(Protocol.IPIP):  'ipip',
	Protocol(Protocol.PIM):   'pim',
	Protocol(Protocol.SCTP):  'sctp',
}

Protocol.VALUE = {
	'icmp':  Protocol(Protocol.ICMP),
	'igmp':  Protocol(Protocol.IGMP),
	'tcp':   Protocol(Protocol.TCP),
	'egp':   Protocol(Protocol.EGP),
	'udp':   Protocol(Protocol.UDP),
	'rsvp':  Protocol(Protocol.RSVP),
	'gre':   Protocol(Protocol.GRE),
	'esp':   Protocol(Protocol.ESP),
	'ah':    Protocol(Protocol.AH),
	'ospf':  Protocol(Protocol.OSPF),
	'ipip':  Protocol(Protocol.IPIP),
	'pim':   Protocol(Protocol.PIM),
	'sctp':  Protocol(Protocol.SCTP),
}



# encoding: utf-8
"""
od.py

Created by Thomas Mangin on 2009-09-06.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""


def od (value):
	def spaced (value):
		even = None
		for v in value:
			if even is False:
				yield ' '
			yield '%02X' % ord(v)
			even = not even
	return ''.join(spaced(value))



# encoding: utf-8
"""
unknown.py

Created by Thomas Mangin on 2013-07-20.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.bgp.message import Message

# ================================================================= UnknownMessage
#


class UnknownMessage (Message):
	# Make sure we have a value, which is not defined in any RFC !

	def __init__ (self, code, data=''):
		self.ID = code
		self.TYPE = chr(code)
		self.data = data

	def message (self):
		return self._message(self.data)

	def __str__ (self):
		return "UNKNOWN"

	@classmethod
	def unpack_message (cls, data):  # pylint: disable=W0613
		raise RuntimeError('should not have been used')

UnknownMessage.klass_unknown = UnknownMessage



# encoding: utf-8
"""
setup.py

Created by Thomas Mangin on 2014-12-23.
Copyright (c) 2014-2015 Exa Networks. All rights reserved.
"""

from exabgp.configuration.environment import environment

_SPACE = {
	'space':  ' '*33
}

HELP_STDOUT = """\
where logging should log
%(space)s syslog (or no setting) sends the data to the local syslog syslog
%(space)s host:<location> sends the data to a remote syslog server
%(space)s stdout sends the data to stdout
%(space)s stderr sends the data to stderr
%(space)s <filename> send the data to a file""" % _SPACE


environment.application = 'exabgp'
environment.configuration = {
	'profile':  {
		'enable':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'false',
			'help':  'toggle profiling of the code',
		},
		'file':  {
			'read':  environment.unquote,
			'write': environment.quote,
			'value': '',
			'help':  'profiling result file, none means stdout, no overwriting',
		},
	},
	'pdb':  {
		'enable':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'false',
			'help':  'on program fault, start pdb the python interactive debugger',
		}
	},
	'daemon':  {
		'pid':  {
			'read':  environment.unquote,
			'write': environment.quote,
			'value': '',
			'help':  'where to save the pid if we manage it',
		},
		'user':  {
			'read':  environment.user,
			'write': environment.quote,
			'value': 'nobody',
			'help':  'user to run as',
		},
		'daemonize':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'false',
			'help':  'should we run in the background',
		},
		'drop':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'true',
			'help':  'drop privileges before forking processes',
		},
	},
	'log':  {
		'enable':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'true',
			'help':  'enable logging',
		},
		'level':  {
			'read':  environment.syslog_value,
			'write': environment.syslog_name,
			'value': 'INFO',
			'help':  'log message with at least the priority SYSLOG.<level>',
		},
		'destination':  {
			'read':  environment.unquote,
			'write': environment.quote,
			'value': 'stdout',
			'help':  HELP_STDOUT,
		},
		'all':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'false',
			'help':  'report debug information for everything',
		},
		'configuration':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'true',
			'help':  'report command parsing',
		},
		'reactor':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'true',
			'help':  'report signal received, command reload',
		},
		'daemon':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'true',
			'help':  'report pid change, forking, ...',
		},
		'processes':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'true',
			'help':  'report handling of forked processes',
		},
		'network':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'true',
			'help':  'report networking information (TCP/IP, network state,...)',
		},
		'packets':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'false',
			'help':  'report BGP packets sent and received',
		},
		'rib':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'false',
			'help':  'report change in locally configured routes',
		},
		'message':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'false',
			'help':  'report changes in route announcement on config reload',
		},
		'timers':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'false',
			'help':  'report keepalives timers',
		},
		'routes':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'false',
			'help':  'report received routes',
		},
		'parser':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'false',
			'help':  'report BGP message parsing details',
		},
		'short':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'false',
			'help':  'use short log format (not prepended with time,level,pid and source)',
		},
	},
	'tcp':  {
		'once': {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'false',
			'help':  'only one tcp connection attempt per peer (for debuging scripts)',
		},
		'delay': {
			'read':  environment.integer,
			'write': environment.nop,
			'value': '0',
			'help':  'start to announce route when the minutes in the hours is a modulo of this number',
		},
		'bind': {
			'read':  environment.optional_ip,
			'write': environment.quote,
			'value': '',
			'help':  'IP to bind on when listening (no ip to disable)',
		},
		'port': {
			'read':  environment.integer,
			'write': environment.nop,
			'value': '179',
			'help':  'port to bind on when listening',
		},
		'acl': {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': '',
			'help':  '(experimental) unimplemented',
		},
	},
	'bgp':  {
		'openwait': {
			'read':  environment.integer,
			'write': environment.nop,
			'value': '60',
			'help':  'how many second we wait for an open once the TCP session is established',
		},
	},
	'cache':  {
		'attributes':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'true',
			'help':  'cache all attributes (configuration and wire) for faster parsing',
		},
		'nexthops':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'true',
			'help':  'cache routes next-hops (deprecated: next-hops are always cached)',
		},
	},
	'api':  {
		'encoder':  {
			'read':  environment.api,
			'write': environment.lower,
			'value': 'text',
			'help':  '(experimental) default encoder to use with with external API (text or json)',
		},
		'highres':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'false',
			'help':  'should we use highres timer in JSON',
		},
		'respawn':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'false',
			'help':  'should we respawn a helper process if it dies',
		},
		'socket':  {
			'read':  environment.unquote,
			'write': environment.quote,
			'value': '',
			'help':  'where should we create a socket for remote control',
		},
	},
	'reactor':  {
		'speed':  {
			'read':  environment.real,
			'write': environment.nop,
			'value': '1.0',
			'help':  'reactor loop time\n%(space)s use only if you understand the code.' % _SPACE,
		},
	},
	# Here for internal use
	'debug':  {
		'pdb':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'false',
			'help':  'enable python debugger on errors',
		},
		'memory':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'false',
			'help':  'command line option --memory',
		},
		'configuration':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'false',
			'help':  'undocumented option: raise when parsing configuration errors',
		},
		'selfcheck':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'false',
			'help':  'does a self check on the configuration file',
		},
		'route':  {
			'read':  environment.unquote,
			'write': environment.quote,
			'value': '',
			'help':  'decode the route using the configuration',
		},
		'defensive':  {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'false',
			'help':  'generate random fault in the code in purpose',
		},
		'rotate': {
			'read':  environment.boolean,
			'write': environment.lower,
			'value': 'false',
			'help':  'rotate configurations file on reload (signal)',
		},
	},
}



# encoding: utf-8
"""
json.py

Created by Thomas Mangin on 2009-08-25.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

# from exabgp.configuration.ancient import Configuration
# from exabgp.configuration.format import formated
#
# from exabgp.protocol.family import AFI
# from exabgp.protocol.family import SAFI
# from exabgp.protocol.family import Family
# from exabgp.protocol.ip import IP
# from exabgp.bgp.message import OUT
#
# from exabgp.bgp.message.update.nlri.prefix import Prefix
# from exabgp.bgp.message.update.nlri.mpls import MPLS
# from exabgp.bgp.message.refresh import RouteRefresh
# from exabgp.bgp.message.operational import Advisory
# from exabgp.bgp.message.operational import Query
# from exabgp.bgp.message.operational import Response
#
# from exabgp.rib.change import Change
# from exabgp.version import version
# from exabgp.logger import Logger

import pprint as pp

from exabgp.configuration import Configuration

from exabgp.configuration.show import SectionShow


def parse (command):
	conf = Configuration()
	conf.register(SectionShow,['show'])
	return conf.parse_string(command)

SHOWS = [
	'show { version }',
]

for show in SHOWS:
	parsed = parse(show)
	print '--'
	print 'command:',show
	for section in parsed.keys():
		d = parsed[section]
		for k,v in d.items():
			print 'section %s name %s ' % (section,k)
			pp.pprint(v)
			print
		print



# encoding: utf-8
"""
command.py

Created by Thomas Mangin on 2015-12-15.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

import socket

from exabgp.version import version as _version


def _extract_neighbors (command):
	"""return a list of neighbor definition : the neighbor definition is a list of string which are in the neighbor indexing string"""
	# This function returns a list and a string
	# The first list contains parsed neighbor to match against our defined peers
	# The string is the command to be run for those peers
	# The parsed neighbor is a list of the element making the neighbor string so each part can be checked against the neighbor name

	returned = []
	neighbor,remaining = command.split(' ',1)
	if neighbor != 'neighbor':
		return [],command

	ip,command = remaining.split(' ',1)
	definition = ['neighbor %s' % (ip)]

	while True:
		try:
			key,value,remaining = command.split(' ',2)
		except ValueError:
			key,value = command.split(' ',1)
		if key == ',':
			returned.append(definition)
			_,command = command.split(' ',1)
			definition = []
			continue
		if key not in ['neighbor','local-ip','local-as','peer-as','router-id','family-allowed']:
			if definition:
				returned.append(definition)
			break
		definition.append('%s %s' % (key,value))
		command = remaining

	return returned,command


def shutdown (self, reactor, service, command):
	reactor.api_shutdown()
	reactor.answer(service,'shutdown in progress')
	return True


def reload (self, reactor, service, command):
	reactor.api_reload()
	reactor.answer(service,'reload in progress')
	return True


def restart (self, reactor, service, command):
	reactor.api_restart()
	reactor.answer(service,'restart in progress')
	return True


def version (self, reactor, service, command):
	reactor.answer(service,'exabgp %s' % _version)
	return True

def log (self, reactor, service, command):
	self.logger.processes(command.lstrip().lstrip('#').strip())
	return True

def teardown (self, reactor, service, command):
	try:
		descriptions,command = _extract_neighbors(command)
		_,code = command.split(' ',1)
		for key in reactor.peers:
			for description in descriptions:
				if reactor.match_neighbor(description,key):
					reactor.peers[key].teardown(int(code))
					self.logger.reactor('teardown scheduled for %s' % ' '.join(description))
		return True
	except ValueError:
		return False
	except IndexError:
		return False


def show_neighbor (self, reactor, service, command):
	def callback ():
		for key in reactor.configuration.neighbor.keys():
			neighbor = reactor.configuration.neighbor.get(key, None)
			if not neighbor:
				continue
			for line in str(neighbor).split('\n'):
				reactor.answer(service,line)
				yield True

	reactor.plan(callback())
	return True


def show_neighbors (self, reactor, service, command):
	def callback ():
		for key in reactor.configuration.neighbor.keys():
			neighbor = reactor.configuration.neighbor.get(key, None)
			if not neighbor:
				continue
			for line in str(neighbor).split('\n'):
				reactor.answer(service,line)
				yield True

	reactor.plan(callback())
	return True


def show_neighbor_status (self, reactor, service, command):
	def callback ():
		for peer_name in reactor.peers.keys():
			peer = reactor.peers.get(peer_name, None)
			if not peer:
				continue
			detailed_status = peer.detailed_link_status()
			families = peer.negotiated_families()
			if families:
				families = "negotiated %s" % families
			reactor.answer(service, "%s %s state %s" % (peer_name, families, detailed_status))
			yield True
		reactor.answer(service,"done")

	reactor.plan(callback())
	return True


def show_routes (self, reactor, service, command):
	def callback ():
		last = command.split()[-1]
		if last == 'routes':
			neighbors = reactor.configuration.neighbor.keys()
		else:
			neighbors = [n for n in reactor.configuration.neighbor.keys() if 'neighbor %s' % last in n]
		for key in neighbors:
			neighbor = reactor.configuration.neighbor.get(key, None)
			if not neighbor:
				continue
			for change in list(neighbor.rib.outgoing.sent_changes()):
				reactor.answer(service,'neighbor %s %s' % (neighbor.peer_address,str(change.nlri)))
				yield True

	reactor.plan(callback())
	return True


def show_routes_extensive (self, reactor, service, command):
	def callback ():
		last = command.split()[-1]
		if last == 'extensive':
			neighbors = reactor.configuration.neighbor.keys()
		else:
			neighbors = [n for n in reactor.configuration.neighbor.keys() if 'neighbor %s' % last in n]
		for key in neighbors:
			neighbor = reactor.configuration.neighbor.get(key, None)
			if not neighbor:
				continue
			for change in list(neighbor.rib.outgoing.sent_changes()):
				reactor.answer(service,'neighbor %s %s' % (neighbor.name(),change.extensive()))
				yield True

	reactor.plan(callback())
	return True



def announce_watchdog (self, reactor, service, command):
	def callback (name):
		for key in reactor.configuration.neighbor.keys():
			neighbor = reactor.configuration.neighbor.get(key, None)
			if not neighbor:
				continue
			neighbor.rib.outgoing.announce_watchdog(name)
			yield False
		reactor.route_update = True

	try:
		name = command.split(' ')[2]
	except IndexError:
		name = service
	reactor.plan(callback(name))
	return True


def withdraw_watchdog (self, reactor, service, command):
	def callback (name):
		for key in reactor.configuration.neighbor.keys():
			neighbor = reactor.configuration.neighbor.get(key, None)
			if not neighbor:
				continue
			neighbor.rib.outgoing.withdraw_watchdog(name)
			yield False
		reactor.route_update = True
	try:
		name = command.split(' ')[2]
	except IndexError:
		name = service
	reactor.plan(callback(name))
	return True


def flush_route (self, reactor, service, command):
	def callback (self, peers):
		self.logger.reactor("Flushing routes for %s" % ', '.join(peers if peers else []) if peers is not None else 'all peers')
		for peer_name in peers:
			peer = reactor.peers.get(peer_name, None)
			if not peer:
				continue
			peer.send_new(update=True)
			yield False

	try:
		descriptions,command = _extract_neighbors(command)
		peers = reactor.match_neighbors(descriptions)
		if not peers:
			self.logger.reactor('no neighbor matching the command : %s' % command,'warning')
			return False
		reactor.plan(callback(self,peers))
		return True
	except ValueError:
		return False
	except IndexError:
		return False


def announce_route (self, reactor, service, command):
	def callback (self, command, nexthops):
		changes = self.format.parse_api_route(command,nexthops,'announce')
		if not changes:
			self.logger.reactor("Command could not parse route in : %s" % command,'warning')
			yield True
		else:
			peers = []
			for (peer,change) in changes:
				peers.append(peer)
				reactor.configuration.change_to_peers(change,[peer,])
				self.logger.reactor("Route added to %s : %s" % (', '.join(peers if peers else []) if peers is not None else 'all peers',change.extensive()))
				yield False
			reactor.route_update = True

	try:
		descriptions,command = _extract_neighbors(command)
		peers = reactor.match_neighbors(descriptions)
		if not peers:
			self.logger.reactor('no neighbor matching the command : %s' % command,'warning')
			return False
		reactor.plan(callback(self,command,reactor.nexthops(peers)))
		return True
	except ValueError:
		return False
	except IndexError:
		return False
	except socket.error:
		return False


def withdraw_route (self, reactor, service, command):
	def callback (self, command, nexthops):
		changes = self.format.parse_api_route(command,nexthops,'withdraw')
		if not changes:
			self.logger.reactor("Command could not parse route in : %s" % command,'warning')
			yield True
		else:
			for (peer,change) in changes:
				if reactor.configuration.change_to_peers(change,[peer,]):
					self.logger.reactor("Route removed : %s" % change.extensive())
					yield False
				else:
					self.logger.reactor("Could not find therefore remove route : %s" % change.extensive(),'warning')
					yield False
			reactor.route_update = True

	try:
		descriptions,command = _extract_neighbors(command)
		peers = reactor.match_neighbors(descriptions)
		if not peers:
			self.logger.reactor('no neighbor matching the command : %s' % command,'warning')
			return False
		reactor.plan(callback(self,command,reactor.nexthops(peers)))
		return True
	except ValueError:
		return False
	except IndexError:
		return False
	except socket.error:
		return False


def announce_vpls (self, reactor, service, command):
	def callback (self, command, nexthops):
		changes = self.format.parse_api_vpls(command,nexthops,'announce')
		if not changes:
			self.logger.reactor("Command could not parse vpls in : %s" % command,'warning')
			yield True
		else:
			peers = []
			for (peer,change) in changes:
				peers.append(peer)
				reactor.configuration.change_to_peers(change,[peer,])
				self.logger.reactor("vpls added to %s : %s" % (', '.join(peers if peers else []) if peers is not None else 'all peers',change.extensive()))
				yield False
			reactor.route_update = True

	try:
		descriptions,command = _extract_neighbors(command)
		peers = reactor.match_neighbors(descriptions)
		if not peers:
			self.logger.reactor('no neighbor matching the command : %s' % command,'warning')
			return False
		reactor.plan(callback(self,command,reactor.nexthops(peers)))
		return True
	except ValueError:
		return False
	except IndexError:
		return False
	except socket.error:
		return False


def withdraw_vpls (self, reactor, service, command):
	def callback (self, command, nexthops):
		changes = self.format.parse_api_vpls(command,nexthops,'withdraw')
		if not changes:
			self.logger.reactor("Command could not parse vpls in : %s" % command,'warning')
			yield True
		else:
			for (peer,change) in changes:
				if reactor.configuration.change_to_peers(change,[peer,]):
					self.logger.reactor("vpls removed : %s" % change.extensive())
					yield False
				else:
					self.logger.reactor("Could not find therefore remove vpls : %s" % change.extensive(),'warning')
					yield False
			reactor.route_update = True

	try:
		descriptions,command = _extract_neighbors(command)
		peers = reactor.match_neighbors(descriptions)
		if not peers:
			self.logger.reactor('no neighbor matching the command : %s' % command,'warning')
			return False
		reactor.plan(callback(self,command,reactor.nexthops(peers)))
		return True
	except ValueError:
		return False
	except IndexError:
		return False
	except socket.error:
		return False


def announce_attribute (self, reactor, service, command):
	def callback (self, command, nexthops):
		changes = self.format.parse_api_attribute(command,nexthops,'announce')
		if not changes:
			self.logger.reactor("Command could not parse attribute in : %s" % command,'warning')
			yield True
		else:
			for (peers,change) in changes:
				reactor.configuration.change_to_peers(change,peers)
				self.logger.reactor("Route added to %s : %s" % (', '.join(peers if peers else []) if peers is not None else 'all peers',change.extensive()))
			yield False
			reactor.route_update = True

	try:
		descriptions,command = _extract_neighbors(command)
		peers = reactor.match_neighbors(descriptions)
		if not peers:
			self.logger.reactor('no neighbor matching the command : %s' % command,'warning')
			return False
		reactor.plan(callback(self,command,reactor.nexthops(peers)))
		return True
	except ValueError:
		return False
	except IndexError:
		return False
	except socket.error:
		return False


def withdraw_attribute (self, reactor, service, command):
	def callback (self, command, nexthops):
		changes = self.format.parse_api_attribute(command,nexthops,'withdraw')
		if not changes:
			self.logger.reactor("Command could not parse attribute in : %s" % command,'warning')
			yield True
		else:
			for (peers,change) in changes:
				if reactor.configuration.change_to_peers(change,peers):
					self.logger.reactor("Route removed : %s" % change.extensive())
					yield False
				else:
					self.logger.reactor("Could not find therefore remove route : %s" % change.extensive(),'warning')
					yield False
			reactor.route_update = True

	try:
		descriptions,command = _extract_neighbors(command)
		peers = reactor.match_neighbors(descriptions)
		if not peers:
			self.logger.reactor('no neighbor matching the command : %s' % command,'warning')
			return False
		reactor.plan(callback(self,command,reactor.nexthops(peers)))
		return True
	except ValueError:
		return False
	except IndexError:
		return False
	except socket.error:
		return False


def announce_flow (self, reactor, service, command):
	def callback (self, command, peers):
		changes = self.format.parse_api_flow(command,'announce')
		if not changes:
			self.logger.reactor("Command could not parse flow in : %s" % command)
			yield True
		else:
			for change in changes:
				reactor.configuration.change_to_peers(change,peers)
				self.logger.reactor("Flow added to %s : %s" % (', '.join(peers if peers else []) if peers is not None else 'all peers',change.extensive()))
				yield False
			reactor.route_update = True

	try:
		descriptions,command = _extract_neighbors(command)
		peers = reactor.match_neighbors(descriptions)
		if not peers:
			self.logger.reactor('no neighbor matching the command : %s' % command,'warning')
			return False
		reactor.plan(callback(self,command,peers))
		return True
	except ValueError:
		return False
	except IndexError:
		return False
	except socket.error:
		return False


def withdraw_flow (self, reactor, service, command):
	def callback (self, command, peers):
		changes = self.format.parse_api_flow(command,'withdraw')
		if not changes:
			self.logger.reactor("Command could not parse flow in : %s" % command)
			yield True
		else:
			for change in changes:
				if reactor.configuration.change_to_peers(change,peers):
					self.logger.reactor("Flow found and removed : %s" % change.extensive())
					yield False
				else:
					self.logger.reactor("Could not find therefore remove flow : %s" % change.extensive(),'warning')
					yield False
			reactor.route_update = True

	try:
		descriptions,command = _extract_neighbors(command)
		peers = reactor.match_neighbors(descriptions)
		if not peers:
			self.logger.reactor('no neighbor matching the command : %s' % command,'warning')
			return False
		reactor.plan(callback(self,command,peers))
		return True
	except ValueError:
		return False
	except IndexError:
		return False
	except socket.error:
		return False


def announce_eor (self, reactor, service, command):
	def callback (self, command, peers):
		family = self.format.parse_api_eor(command)
		if not family:
			self.logger.reactor("Command could not parse eor : %s" % command)
			yield True
		else:
			reactor.configuration.eor_to_peers(family,peers)
			self.logger.reactor("Sent to %s : %s" % (', '.join(peers if peers else []) if peers is not None else 'all peers',family.extensive()))
			yield False
			reactor.route_update = True

	try:
		descriptions,command = _extract_neighbors(command)
		peers = reactor.match_neighbors(descriptions)
		if not peers:
			self.logger.reactor('no neighbor matching the command : %s' % command,'warning')
			return False
		reactor.plan(callback(self,command,peers))
		return True
	except ValueError:
		return False
	except IndexError:
		return False


def announce_refresh (self, reactor, service, command):
	def callback (self, command, peers):
		refresh = self.format.parse_api_refresh(command)
		if not refresh:
			self.logger.reactor("Command could not parse flow in : %s" % command)
			yield True
		else:
			reactor.configuration.refresh_to_peers(refresh,peers)
			self.logger.reactor("Sent to %s : %s" % (', '.join(peers if peers else []) if peers is not None else 'all peers',refresh.extensive()))
			yield False
			reactor.route_update = True

	try:
		descriptions,command = _extract_neighbors(command)
		peers = reactor.match_neighbors(descriptions)
		if not peers:
			self.logger.reactor('no neighbor matching the command : %s' % command,'warning')
			return False
		reactor.plan(callback(self,command,peers))
		return True
	except ValueError:
		return False
	except IndexError:
		return False


def announce_operational (self, reactor, service, command):
	def callback (self, command, peers):
		operational = self.format.parse_api_operational(command)
		if not operational:
			self.logger.reactor("Command could not parse operational command : %s" % command)
			yield True
		else:
			reactor.configuration.operational_to_peers(operational,peers)
			self.logger.reactor("operational message sent to %s : %s" % (
				', '.join(peers if peers else []) if peers is not None else 'all peers',operational.extensive()
				)
			)
			yield False
			reactor.route_update = True

	if (command.split() + ['safe'])[1].lower() not in ('asm','adm','rpcq','rpcp','apcq','apcp','lpcq','lpcp'):
		return False

	try:
		descriptions,command = _extract_neighbors(command)
		peers = reactor.match_neighbors(descriptions)
		if not peers:
			self.logger.reactor('no neighbor matching the command : %s' % command,'warning')
			return False
		reactor.plan(callback(self,command,peers))
		return True
	except ValueError:
		return False
	except IndexError:
		return False



# encoding: utf-8
"""
registry.py

Created by Thomas Mangin on 2014-06-22.
Copyright (c) 2014-2015 Exa Networks. All rights reserved.
"""

from exabgp.configuration.engine.raised import Raised
from exabgp.configuration.engine.section import Section

from exabgp.configuration.engine.reader import Reader
from exabgp.configuration.engine.tokeniser import Tokeniser

from collections import defaultdict
from exabgp.util.dictionary import Dictionary

from StringIO import StringIO
import time


# ===================================================================== Registry
# The class where all configuration callback are registered to

class Configuration (object):
	def __init__ (self):
		self.stack = []
		self._klass = {}
		self._handler = {}
		self._parser = None
		self.section = None

	# self.location set by Registry

	def register (self, cls, location):
		cls.register(self,location)

	def register_class (self, cls):
		print
		print "class %s" % cls.__name__
		print "-"*40
		if not cls in self._klass:
			self._klass[cls] = cls()
		print

	def register_hook (self, cls, action, position, function):
		key = '/'.join(position)
		if action in self._handler:
			raise Exception('conflicting handlers')
		self._handler.setdefault(key,{})[action] = getattr(cls,function)
		print "%-50s %-7s %s.%s" % (key if key else 'root',action,cls.__name__,function)

	def iterate (self, tokeniser):
		# each section can registered named configuration for reference here
		Section.configuration[tokeniser.name] = defaultdict(Dictionary)

		def run (search, section, location):
			key = '/'.join(search)
			function = self._handler.get(key,{}).get(section,None)

			if function:
				print 'hit %s/%s' % (key,section)
				instance = self._klass.setdefault(function.im_class,function.im_class())
				instance.location = location
				return function(instance,tokeniser) is None
			return False

		while True:
			token = tokeniser()
			if not token: break

			location = self.stack+[token,]

			# if we have both a section and a action, try the action first
			if run(location,'action',location):
				yield None
				continue

			if run(location,'enter',self.stack):
				self.stack.append(token)
				yield None
				continue

			if token != '}':
				print
				print 'Available paths are .....'
				print
				for path in sorted(self._handler):
					for action in sorted(self._handler[path]):
						print '/%-50s %s' % (path,action)
				print '....'
				print
				print '/'.join(location)
				# we need the line and position at this level
				raise Raised(tokeniser,'no parser for the location /%s' % ('/'.join(location)))

			if run(self.stack,'exit',self.stack[:-1]):
				self.stack.pop()
				yield None
				continue

			# we need the line and position at this level
			raise Exception('application error, no exit code registered for %s, please report with your configuration' % '/'.join(self.stack))

		data = Section.configuration[tokeniser.name]
		del Section.configuration[tokeniser.name]
		yield data

	def parse_tokeniser (self, tokeniser):
		if self._parser is None:
			self._parser = self.iterate(tokeniser)

		next = self._parser.next()
		if next is None:
			return None
		self._parser = None
		return next

	def parse_file (self, fname):
		with Reader(fname) as r:
			tokeniser = Tokeniser('configuration',r)
			parsed = None
			while parsed is None:
				parsed = self.parse_tokeniser(tokeniser)
		return parsed

	def parse_string (self, string):
		name = 'command-%d' % int(int(time.time()*1000) % (365*24*60*60*1000))
		sio = StringIO(string)
		tokeniser = Tokeniser(name,sio)
		parsed = None
		while parsed is None:
			parsed = self.parse_tokeniser(tokeniser)
		return parsed



# encoding: utf-8
"""
aggregator.py

Created by Thomas Mangin on 2012-07-14.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.bgp.message.open.asn import ASN
from exabgp.protocol.ip import IPv4

from exabgp.bgp.message.update.attribute.attribute import Attribute


# =============================================================== AGGREGATOR (7)
#

class Aggregator (Attribute):
	ID = Attribute.CODE.AGGREGATOR
	FLAG = Attribute.Flag.TRANSITIVE | Attribute.Flag.OPTIONAL
	CACHING = True

	__slots__ = ['asn','speaker','_str']

	def __init__ (self, asn, speaker):
		self.asn = asn
		self.speaker = speaker
		self._str = None

	def pack (self, negotiated):
		if negotiated.asn4:
			return self._attribute(self.asn.pack(True)+self.speaker.pack())
		elif self.asn.asn4():
			return self._attribute(self.asn.trans().pack()+self.speaker.pack()) + Aggregator4(self.asn,self.speaker).pack(negotiated)
		else:
			return self._attribute(self.asn.pack()+self.speaker.pack())

	def __len__ (self):
		raise RuntimeError('size can be 6 or 8 - we can not say - or can we ?')

	def __str__ (self):
		if not self._str:
			self._str = '%s:%s' % (self.asn,self.speaker)
		return self._str

	def json (self):
		return '{ "asn" : %d, "speaker" : "%d" }' % (self.asn,self.speaker)

	@classmethod
	def unpack (cls, data, negotiated):
		if negotiated.asn4:
			return cls(ASN.unpack(data[:4]),IPv4.unpack(data[-4:]))
		return cls(ASN.unpack(data[:2]),IPv4.unpack(data[-4:]))


# ============================================================== AGGREGATOR (18)
#

class Aggregator4 (Aggregator):
	ID = Attribute.CODE.AS4_AGGREGATOR
	__slots__ = ['pack']

	def pack (self, negotiated):
		return self._attribute(self.asn.pack(True)+self.speaker.pack())



# encoding: utf-8
"""
daemon.py

Created by Thomas Mangin on 2011-05-02.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

import os
import sys
import pwd
import errno
import socket

from exabgp.configuration.environment import environment

from exabgp.logger import Logger

MAXFD = 2048


class Daemon (object):

	def __init__ (self, reactor):
		self.pid = environment.settings().daemon.pid
		self.user = environment.settings().daemon.user
		self.daemonize = environment.settings().daemon.daemonize

		self.logger = Logger()

		self.reactor = reactor

		os.chdir('/')
		# os.umask(0)
		os.umask(0137)

	def check_pid (self,pid):
		if pid < 0:  # user input error
			return False
		if pid == 0:  # all processes
			return False
		try:
			os.kill(pid, 0)
			return True
		except OSError as err:
			if err.errno == errno.EPERM:  # a process we were denied access to
				return True
			if err.errno == errno.ESRCH:  # No such process
				return False
			# should never happen
			return False

	def savepid (self):
		self._saved_pid = False

		if not self.pid:
			return True

		ownid = os.getpid()

		flags = os.O_CREAT | os.O_EXCL | os.O_WRONLY
		mode = ((os.R_OK | os.W_OK) << 6) | (os.R_OK << 3) | os.R_OK

		try:
			fd = os.open(self.pid,flags,mode)
		except OSError:
			try:
				pid = open(self.pid,'r').readline().strip()
				if self.check_pid(int(pid)):
					self.logger.daemon("PIDfile already exists and program still running %s" % self.pid)
					return False
				else:
					# If pid is not running, reopen file without O_EXCL
					fd = os.open(self.pid,flags ^ os.O_EXCL,mode)
			except (OSError,IOError,ValueError):
				self.logger.daemon("issue accessing PID file %s (most likely permission or ownership)" % self.pid)
				return False

		try:
			f = os.fdopen(fd,'w')
			line = "%d\n" % ownid
			f.write(line)
			f.close()
			self._saved_pid = True
		except IOError:
			self.logger.daemon("Can not create PIDfile %s" % self.pid,'warning')
			return False
		self.logger.daemon("Created PIDfile %s with value %d" % (self.pid,ownid),'warning')
		return True

	def removepid (self):
		if not self.pid or not self._saved_pid:
			return
		try:
			os.remove(self.pid)
		except OSError,exc:
			if exc.errno == errno.ENOENT:
				pass
			else:
				self.logger.daemon("Can not remove PIDfile %s" % self.pid,'error')
				return
		self.logger.daemon("Removed PIDfile %s" % self.pid)

	def drop_privileges (self):
		"""return true if we are left with insecure privileges"""
		# os.name can be ['posix', 'nt', 'os2', 'ce', 'java', 'riscos']
		if os.name not in ['posix',]:
			return True

		uid = os.getuid()
		gid = os.getgid()

		if uid and gid:
			return True

		try:
			user = pwd.getpwnam(self.user)
			nuid = int(user.pw_uid)
			ngid = int(user.pw_gid)
		except KeyError:
			return False

		# not sure you can change your gid if you do not have a pid of zero
		try:
			# we must change the GID first otherwise it may fail after change UID
			if not gid:
				os.setgid(ngid)
			if not uid:
				os.setuid(nuid)

			cuid = os.getuid()
			ceid = os.geteuid()
			cgid = os.getgid()

			if cuid < 0:
				cuid += (1 << 32)

			if cgid < 0:
				cgid += (1 << 32)

			if ceid < 0:
				ceid += (1 << 32)

			if nuid != cuid or nuid != ceid or ngid != cgid:
				return False

		except OSError:
			return False

		return True

	def _is_socket (self, fd):
		try:
			s = socket.fromfd(fd, socket.AF_INET, socket.SOCK_RAW)
		except ValueError:
			# The file descriptor is closed
			return False
		try:
			s.getsockopt(socket.SOL_SOCKET, socket.SO_TYPE)
		except socket.error,exc:
			# It is look like one but it is not a socket ...
			if exc.args[0] == errno.ENOTSOCK:
				return False
		return True

	def daemonise (self):
		if not self.daemonize:
			return

		log = environment.settings().log
		if log.enable and log.destination.lower() in ('stdout','stderr'):
			self.logger.daemon('ExaBGP can not fork when logs are going to %s' % log.destination.lower(),'critical')
			return

		def fork_exit ():
			try:
				pid = os.fork()
				if pid > 0:
					os._exit(0)
			except OSError,exc:
				self.logger.reactor('Can not fork, errno %d : %s' % (exc.errno,exc.strerror),'critical')

		# do not detach if we are already supervised or run by init like process
		if self._is_socket(sys.__stdin__.fileno()) or os.getppid() == 1:
			return

		fork_exit()
		os.setsid()
		fork_exit()
		self.silence()

	def silence (self):
		# closing more would close the log file too if open
		maxfd = 3

		for fd in range(0, maxfd):
			try:
				os.close(fd)
			except OSError:
				pass
		os.open("/dev/null", os.O_RDWR)
		os.dup2(0, 1)
		os.dup2(0, 2)

		# import resource
		# if 'linux' in sys.platform:
		# 	nofile = resource.RLIMIT_NOFILE
		# elif 'bsd' in sys.platform:
		# 	nofile = resource.RLIMIT_OFILE
		# else:
		# 	self.logger.daemon("For platform %s, can not close FDS before forking" % sys.platform)
		# 	nofile = None
		# if nofile:
		# 	maxfd = resource.getrlimit(nofile)[1]
		# 	if (maxfd == resource.RLIM_INFINITY):
		# 		maxfd = MAXFD
		# else:
		# 	maxfd = MAXFD



#!/usr/bin/env python
# encoding: utf-8
"""
cli.py

Created by Thomas Mangin on 2014-12-22.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

import sys
from exabgp.dep.cmd2 import cmd

from exabgp.version import version


class Completed (cmd.Cmd):
	# use_rawinput = False
	# prompt = ''

	# doc_header = 'doc_header'
	# misc_header = 'misc_header'
	# undoc_header = 'undoc_header'

	ruler = '-'
	completion = {}

	def __init__ (self, intro=''):
		self.prompt = '%s> ' % intro
		cmd.Cmd.__init__(self)

	def completedefault (self, text, line, begidx, endidx):  # pylint: disable=W0613
		commands = line.split()
		local = self.completion

		for command in commands:
			if command in local:
				local = local[command]
				continue
			break

		return [_ for _ in local.keys() if _.startswith(text)]

	def default (self, line):
		print 'unrecognised syntax: ', line

	def do_EOF (self):
		return True


class SubMenu (Completed):
	def do_exit (self, _):
		return True

	do_x = do_exit


class Attribute (SubMenu):
	chars = ''.join(chr(_) for _ in range(ord('a'),ord('z')+1) + range(ord('0'),ord('9')+1) + [ord ('-')])

	attribute = None

	completion = {
		'origin':  {
			'igp': {
			},
			'egp': {
			},
			'incomplete': {
			},
		},
	}

	def __init__ (self, name):
		self.name = name
		SubMenu.__init__(self,'attribute %s' % name)

	def do_origin (self, line):
		if line in ('igp','egp','incomplete'):
			self.attribute['origin'] = line
		else:
			print 'invalid origin'

	def do_as_path (self, line):
		pass

	# next-hop

	def do_med (self, line):
		if not line.isdigit():
			print 'invalid med, %s is not a number' % line
			return

		med = int(line)
		if 0 > med < 65536:
			print 'invalid med, %s is not a valid number' % line
		self.attribute['origin'] = line

	# local-preference
	# atomic-aggregate
	# aggregator
	# community
	# originator-id
	# cluster-list
	# extended-community
	# psmi
	# aigp

	def do_show (self, _):
		print 'attribute %s ' % self.name + ' '.join('%s %s' % (key,value) for key,value in self.attribute.iteritems())


class ExaBGP (Completed):
	completion = {
		'announce':  {
			'route':  {
			},
			'l2vpn':  {
			},
		},
		'neighbor': {
			'include': {
			},
			'exclude': {
			},
			'reset': {
			},
			'list': {
			},
		},
		'attribute':  {
		},
		'show': {
			'routes':  {
				'extensive': {
				},
				'minimal': {
				},
			},
		},
		'reload': {
		},
		'restart': {
		},
	}

	def _update_prompt (self):
		if self._neighbors:
			self.prompt = '\n# neighbor ' + ', '.join(self._neighbors) + '\n> '
		else:
			self.prompt = '\n> '

	#
	# repeat last command
	#

	# last = 'help'

	# def do_last (self, line):
	# 	"Print the input, replacing '$out' with the output of the last shell command"
	# 	# Obviously not robust
	# 	if hasattr(self, 'last_output'):
	# 		print line.replace('$out', self.last_output)

	_neighbors = set()

	def do_neighbor (self, line):
		try:
			action,ip = line.split()
		except ValueError:
			if line == 'reset':
				print 'removed neighbors', ', '.join(self._neighbors)
				self._neighbors = set()
				self._update_prompt()
			else:
				print 'invalid syntax'
				self.help_neighbor()
			return

		if action == 'include':
			# check ip is an IP
			# check ip is a known IP
			self._neighbors.add(ip)
			self._update_prompt()
		elif action == 'exclude':
			if ip in self._neighbors:
				self._neighbors.remove(ip)
				print 'neighbor excluded'
				self._update_prompt()
			else:
				print 'invalid neighbor'
		elif action == 'list':
			print 'removed neighbors', ', '.join(self._neighbors)
		else:
			print 'invalid syntax'
			self.help_neighbor()

	def help_neighbor (self):
		print "neighbor include <ip>:  limit the action to the defined neighbors"
		print "neighbor exclude <ip>:  remove a particular neighbor"
		print "neighbor reset       :  clear the neighbor previous set "

	_attribute = {}

	def do_attribute (self, name):
		if not name:
			self.help_attribute()
			return
		invalid = ''.join([_ for _ in name if _ not in Attribute.chars])
		if invalid:
			print 'invalid character(s) in attribute name: %s' % invalid
			return
		cli = Attribute(name)
		cli.attribute = self._attribute.get(name,{})
		cli.cmdloop()

	def help_attribute (self):
		print 'attribute <name>'

	def do_quit (self, _):
		return True

	do_q = do_quit


def main ():
	if len(sys.argv) > 1:
		ExaBGP().onecmd(' '.join(sys.argv[1:]))
	else:
		print "ExaBGP %s CLI" % version
		ExaBGP('').cmdloop()


if __name__ == '__main__':
	main()



# encoding: utf-8
"""
mprnlri.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from struct import unpack

from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI
from exabgp.protocol.ip.address import Address

from exabgp.bgp.message import IN
from exabgp.bgp.message.update.attribute.attribute import Attribute
from exabgp.bgp.message.update.nlri.nlri import NLRI

from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.open.capability.negotiated import Negotiated


# ================================================================= MP NLRI (14)

class MPURNLRI (Attribute,Address):
	FLAG = Attribute.Flag.OPTIONAL
	ID = Attribute.CODE.MP_UNREACH_NLRI

	# __slots__ = ['nlris']

	def __init__ (self, afi, safi, nlris):
		Address.__init__(self,afi,safi)
		self.nlris = nlris

	def packed_attributes (self, addpath, maximum):
		if not self.nlris:
			return

		mpurnlri = {}
		for nlri in self.nlris:
			mpurnlri.setdefault((nlri.afi.pack(),nlri.safi.pack()),[]).append(nlri.pack(addpath))

		for (pafi,psafi),nlris in mpurnlri.iteritems():
			payload = pafi + psafi + ''.join(nlris)

			if self._len(payload) <= maximum:
				yield self._attribute(payload)
				continue

			# This will not generate an optimum update size..
			# we should feedback the maximum on each iteration

			for nlri in nlris:
				yield self._attribute(pafi + psafi + nlri)


	def pack (self, addpath):
		return ''.join(self.packed_attributes(addpath,Negotiated.FREE_SIZE))

	def __len__ (self):
		raise RuntimeError('we can not give you the size of an MPURNLRI - was it with our witout addpath ?')

	def __str__ (self):
		return "MP_UNREACH_NLRI for %s %s with %d NLRI(s)" % (self.afi,self.safi,len(self.nlris))

	@classmethod
	def unpack (cls, data, negotiated):
		nlris = []

		# -- Reading AFI/SAFI
		afi,safi = unpack('!HB',data[:3])
		offset = 3
		data = data[offset:]

		if negotiated and (afi,safi) not in negotiated.families:
			raise Notify(3,0,'presented a non-negotiated family %s %s' % (AFI(afi),SAFI(safi)))

		# Is the peer going to send us some Path Information with the route (AddPath)
		addpath = negotiated.addpath.receive(afi,safi)

		while data:
			length,nlri = NLRI.unpack(afi,safi,data,addpath,None,IN.WITHDRAWN)
			nlris.append(nlri)
			data = data[length:]
			# logger.parser(LazyFormat("parsed withdraw mp nlri %s payload " % nlri,data[:length]))

		return cls(afi,safi,nlris)

EMPTY_MPURNLRI = MPURNLRI(AFI(AFI.undefined),SAFI(SAFI.undefined),[])



# encoding: utf-8
"""
protocol.py

Created by Thomas Mangin on 2009-08-25.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

import os

from exabgp.reactor.network.outgoing import Outgoing
# from exabgp.reactor.network.error import NotifyError

from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI
from exabgp.bgp.message import Message
from exabgp.bgp.message.nop import NOP
from exabgp.bgp.message.nop import _NOP
from exabgp.bgp.message.open import Open
from exabgp.bgp.message.open.capability.capability import Capability
from exabgp.bgp.message.open.capability.capabilities import Capabilities
from exabgp.bgp.message.open.capability.negotiated import Negotiated
from exabgp.bgp.message.update import Update
from exabgp.bgp.message.update import EOR
from exabgp.bgp.message.keepalive import KeepAlive
from exabgp.bgp.message.notification import Notification
from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.operational import Operational

from exabgp.reactor.api.processes import ProcessError

from exabgp.logger import Logger
from exabgp.logger import FakeLogger

# This is the number of chuncked message we are willing to buffer, not the number of routes
MAX_BACKLOG = 15000

_UPDATE = Update([],'')
_OPERATIONAL = Operational(0x00)


class Protocol (object):
	decode = True

	def __init__ (self, peer):
		try:
			self.logger = Logger()
		except RuntimeError:
			self.logger = FakeLogger()
		self.peer = peer
		self.neighbor = peer.neighbor
		self.negotiated = Negotiated(self.neighbor)
		self.connection = None

		if self.neighbor.connect:
			self.port = self.neighbor.connect
		elif os.environ.get('exabgp.tcp.port','').isdigit():
			self.port = int(os.environ.get('exabgp.tcp.port'))
		elif os.environ.get('exabgp_tcp_port','').isdigit():
			self.port = int(os.environ.get('exabgp_tcp_port'))
		else:
			self.port = 179

		# XXX: FIXME: check the the -19 is correct (but it is harmless)
		# The message size is the whole BGP message _without_ headers
		self.message_size = Message.MAX_LEN-Message.HEADER_LEN

		from exabgp.configuration.environment import environment
		self.log_routes = environment.settings().log.routes

	# XXX: we use self.peer.neighbor.peer_address when we could use self.neighbor.peer_address

	# seems to be triggering a bug where the GC calls this but the connection stays up !

	# def __del__ (self):
	# 	self.close('automatic protocol cleanup')

	def me (self, message):
		return "Peer %15s ASN %-7s %s" % (self.peer.neighbor.peer_address,self.peer.neighbor.peer_as,message)

	def accept (self, incoming):
		self.connection = incoming

		if self.peer.neighbor.api['neighbor-changes']:
			self.peer.reactor.processes.connected(self.peer)

		# very important - as we use this function on __init__
		return self

	def connect (self):
		# allows to test the protocol code using modified StringIO with a extra 'pending' function
		if not self.connection:
			peer = self.neighbor.peer_address
			local = self.neighbor.local_address
			md5 = self.neighbor.md5
			ttl = self.neighbor.ttl
			self.connection = Outgoing(peer.afi,peer.ip,local.ip,self.port,md5,ttl)

			try:
				generator = self.connection.establish()
				while True:
					connected = generator.next()
					if not connected:
						yield False
						continue
					if self.peer.neighbor.api['neighbor-changes']:
						self.peer.reactor.processes.connected(self.peer)
					yield True
					return
			except StopIteration:
				# close called by the caller
				# self.close('could not connect to remote end')
				yield False
				return

	def close (self, reason='protocol closed, reason unspecified',notification=True):
		if self.connection:
			self.logger.network(self.me(reason))

			# must be first otherwise we could have a loop caused by the raise in the below
			self.connection.close()
			self.connection = None

			if self.peer.neighbor.api['neighbor-changes'] and notification:
				try:
					self.peer.reactor.processes.down(self.peer,reason)
				except ProcessError:
					self.logger.message(self.me('could not send notification of neighbor close to API'))

	def write (self, message):
		if self.neighbor.api['send-packets'] and not self.neighbor.api['consolidate']:
			self.peer.reactor.processes.send(self.peer,ord(message[18]),message[:19],message[19:])
		for boolean in self.connection.writer(message):
			yield boolean

	# Read from network .......................................................

	def read_message (self):
		# This will always be defined by the loop but scope leaking upset scrutinizer/pylint
		msg_id = None

		for length,msg_id,header,body,notify in self.connection.reader():
			if notify:
				if self.neighbor.api['receive-packets']:
					self.peer.reactor.processes.receive(self.peer,msg_id,header,body)
				if self.neighbor.api[Message.CODE.NOTIFICATION]:
					self.peer.reactor.processes.notification(self.peer,notify.code,notify.subcode,str(notify))
				# XXX: is notify not already Notify class ?
				raise Notify(notify.code,notify.subcode,str(notify))
			if not length:
				yield _NOP

		if self.neighbor.api['receive-packets'] and not self.neighbor.api['consolidate']:
			self.peer.reactor.processes.receive(self.peer,msg_id,header,body)

		if msg_id == Message.CODE.UPDATE and not self.neighbor.api['receive-parsed'] and not self.log_routes:
			yield _UPDATE
			return

		self.logger.message(self.me('<< %s' % Message.CODE.name(msg_id)))
		try:
			message = Message.unpack(msg_id,body,self.negotiated)
		except (KeyboardInterrupt,SystemExit,Notify):
			raise
		except Exception,exc:
			self.logger.message(self.me('Could not decode message "%d"' % msg_id))
			self.logger.message(self.me('%s' % str(exc)))
			# XXX: TODO: add backtrace here
			raise Notify(1,0,'can not decode update message of type "%d"' % msg_id)
			# raise Notify(5,0,'unknown message received')

		if self.neighbor.api[msg_id]:
			if self.neighbor.api['receive-parsed']:
				if self.neighbor.api['consolidate'] and self.neighbor.api['receive-packets']:
					self.peer.reactor.processes.message(msg_id,self.peer,message,header,body)
				else:
					self.peer.reactor.processes.message(msg_id,self.peer,message,'','')

		if message.TYPE == Notification.TYPE:
			raise message

		yield message

		# elif msg == Message.CODE.ROUTE_REFRESH:
		# 	if self.negotiated.refresh != REFRESH.ABSENT:
		# 		self.logger.message(self.me('<< ROUTE-REFRESH'))
		# 		refresh = RouteRefresh.unpack_message(body,self.negotiated)
		# 		if self.neighbor.api.receive_refresh:
		# 			if refresh.reserved in (RouteRefresh.start,RouteRefresh.end):
		# 				if self.neighbor.api.consolidate:
		# 					self.peer.reactor.process.refresh(self.peer,refresh,header,body)
		# 				else:
		# 					self.peer.reactor.processes.refresh(self.peer,refresh,'','')
		# 	else:
		# 		# XXX: FIXME: really should raise, we are too nice
		# 		self.logger.message(self.me('<< NOP (un-negotiated type %d)' % msg))
		# 		refresh = UnknownMessage.unpack_message(body,self.negotiated)
		# 	yield refresh

	def validate_open (self):
		error = self.negotiated.validate(self.neighbor)
		if error is not None:
			raise Notify(*error)

	def read_open (self, ip):
		for received_open in self.read_message():
			if received_open.TYPE == NOP.TYPE:
				yield received_open
			else:
				break

		if received_open.TYPE != Open.TYPE:
			raise Notify(5,1,'The first packet recevied is not an open message (%s)' % received_open)

		self.logger.message(self.me('<< %s' % received_open))
		yield received_open

	def read_keepalive (self):
		for message in self.read_message():
			if message.TYPE == NOP.TYPE:
				yield message
			else:
				break

		if message.TYPE != KeepAlive.TYPE:
			raise Notify(5,2)

		yield message

	#
	# Sending message to peer
	#

	def new_open (self, restarted):
		sent_open = Open(
			4,
			self.neighbor.local_as,
			self.neighbor.router_id.ip,
			Capabilities().new(self.neighbor,restarted),
			self.neighbor.hold_time
		)

		# we do not buffer open message in purpose
		msg_send = sent_open.message()
		for _ in self.write(msg_send):
			yield _NOP

		self.logger.message(self.me('>> %s' % sent_open))
		if self.neighbor.api[Message.CODE.OPEN]:
			if self.neighbor.api['consolidate']:
				header = msg_send[0:38]
				body = msg_send[38:]
				self.peer.reactor.processes.message(Message.CODE.OPEN,self.peer,sent_open,header,body,'sent')
			else:
				self.peer.reactor.processes.message(Message.CODE.OPEN,self.peer,sent_open,'','','sent')
		yield sent_open

	def new_keepalive (self, comment=''):
		keepalive = KeepAlive()

		for _ in self.write(keepalive.message()):
			yield _NOP

		self.logger.message(self.me('>> KEEPALIVE%s' % (' (%s)' % comment if comment else '')))

		yield keepalive

	def new_notification (self, notification):
		for _ in self.write(notification.message()):
			yield _NOP
		self.logger.message(self.me('>> NOTIFICATION (%d,%d,"%s")' % (notification.code,notification.subcode,notification.data)))
		yield notification

	def new_update (self):
		updates = self.neighbor.rib.outgoing.updates(self.neighbor.group_updates)
		number = 0
		for update in updates:
			for message in update.messages(self.negotiated):
				number += 1
				for boolean in self.write(message):
					# boolean is a transient network error we already announced
					yield _NOP
		if number:
			self.logger.message(self.me('>> %d UPDATE(s)' % number))
		yield _UPDATE

	def new_eor (self, afi, safi):
		eor = EOR(afi,safi)
		for _ in self.write(eor.message()):
			yield _NOP
		self.logger.message(self.me('>> EOR %s %s' % (afi,safi)))
		yield eor

	def new_eors (self, afi=AFI.undefined,safi=SAFI.undefined):
		# Send EOR to let our peer know he can perform a RIB update
		if self.negotiated.families:
			families = self.negotiated.families if (afi,safi) == (AFI.undefined,SAFI.undefined) else [(afi,safi),]
			for eor_afi,eor_safi in families:
				for _ in self.new_eor(eor_afi,eor_safi):
					yield _
		else:
			# If we are not sending an EOR, send a keepalive as soon as when finished
			# So the other routers knows that we have no (more) routes to send ...
			# (is that behaviour documented somewhere ??)
			for eor in self.new_keepalive('EOR'):
				yield _NOP
			yield _UPDATE

	def new_operational (self, operational, negotiated):
		for _ in self.write(operational.message(negotiated)):
			yield _NOP
		self.logger.message(self.me('>> OPERATIONAL %s' % str(operational)))
		yield operational

	def new_refresh (self, refresh):
		for _ in self.write(refresh.message()):
			yield _NOP
		self.logger.message(self.me('>> REFRESH %s' % str(refresh)))
		yield refresh



"""
process.py

Created by Thomas Mangin on 2011-05-02.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

import os
import errno
import time
import subprocess
import select
import fcntl

from exabgp.util.errstr import errstr
from exabgp.reactor.network.error import error

from exabgp.configuration.format import formated
from exabgp.reactor.api.encoding import Text
from exabgp.reactor.api.encoding import JSON
from exabgp.bgp.message import Message
from exabgp.logger import Logger


# pylint: disable=no-self-argument,not-callable,unused-argument,invalid-name

class ProcessError (Exception):
	pass


def preexec_helper ():
	# make this process a new process group
	# os.setsid()
	# This prevent the signal to be sent to the children (and create a new process group)
	os.setpgrp()
	# signal.signal(signal.SIGINT, signal.SIG_IGN)


class Processes (object):
	# how many time can a process can respawn in the time interval
	respawn_number = 5
	respawn_timemask = 0xFFFFFF - pow(2,6) + 1
	# '0b111111111111111111000000' (around a minute, 63 seconds)

	_dispatch = {}

	# names = {
	# 	Message.CODE.NOTIFICATION  : 'neighbor-changes',
	# 	Message.CODE.OPEN          : 'receive-opens',
	# 	Message.CODE.KEEPALIVE     : 'receive-keepalives',
	# 	Message.CODE.UPDATE        : 'receive-updates',
	# 	Message.CODE.ROUTE_REFRESH : 'receive-refresh',
	# 	Message.CODE.OPERATIONAL   : 'receive-operational',
	# }

	def __init__ (self, reactor):
		self.logger = Logger()
		self.reactor = reactor
		self.clean()
		self.silence = False

		from exabgp.configuration.environment import environment
		self.highres = environment.settings().api.highres

	def clean (self):
		self._process = {}
		self._encoder = {}
		self._events = {}
		self._neighbor_process = {}
		self._respawning = {}

	def _terminate (self, process):
		self.logger.processes("Terminating process %s" % process)
		try:
			self._process[process].terminate()
		except OSError:
			# the process is most likely already dead
			pass
		self._process[process].wait()
		del self._process[process]

		for processes in self._neighbor_process.itervalues():
			if process in processes:
				processes.remove(process)

	def terminate (self):
		for process in list(self._process):
			if not self.silence:
				try:
					self.write(process,self._encoder[process].shutdown())
				except ProcessError:
					pass
		self.silence = True
		time.sleep(0.1)
		for process in list(self._process):
			try:
				self._terminate(process)
			except OSError:
				# we most likely received a SIGTERM signal and our child is already dead
				self.logger.processes("child process %s was already dead" % process)

		self.clean()

	def _start (self, process):
		events = self.reactor.configuration.process[process]
		for event,present in events.iteritems():
			if event in ('run','encoder'):
				continue
			if present:
				self._events.setdefault(process,[]).append(event)

		try:
			if process in self._process:
				self.logger.processes("process already running")
				return
			if process not in self.reactor.configuration.process:
				self.logger.processes("Can not start process, no configuration for it (anymore ?)")
				return

			# Prevent some weird termcap data to be created at the start of the PIPE
			# \x1b[?1034h (no-eol) (esc)
			os.environ['TERM'] = 'dumb'

			run = self.reactor.configuration.process[process].get('run','')
			if run:
				api = self.reactor.configuration.process[process]['encoder']
				self._encoder[process] = JSON('3.4.8',self.highres) if api == 'json' else Text('3.3.2')

				self._process[process] = subprocess.Popen(
					run,
					stdin=subprocess.PIPE,
					stdout=subprocess.PIPE,
					preexec_fn=preexec_helper
					# This flags exists for python 2.7.3 in the documentation but on on my MAC
					# creationflags=subprocess.CREATE_NEW_PROCESS_GROUP
				)
				fcntl.fcntl(self._process[process].stdout.fileno(), fcntl.F_SETFL, os.O_NONBLOCK)

				self.logger.processes("Forked process %s" % process)

				around_now = int(time.time()) & self.respawn_timemask
				if process in self._respawning:
					if around_now in self._respawning[process]:
						self._respawning[process][around_now] += 1
						# we are respawning too fast
						if self._respawning[process][around_now] > self.respawn_number:
							self.logger.processes(
								"Too many respawn for %s (%d) terminating program" % (process,self.respawn_number),
								'critical'
							)
							raise ProcessError()
					else:
						# reset long time since last respawn
						self._respawning[process] = {around_now: 1}
				else:
					# record respawing
					self._respawning[process] = {around_now: 1}

			neighbor = self.reactor.configuration.process[process]['neighbor']
			self._neighbor_process.setdefault(neighbor,[]).append(process)
		except (subprocess.CalledProcessError,OSError,ValueError),exc:
			self.logger.processes("Could not start process %s" % process)
			self.logger.processes("reason: %s" % str(exc))
			raise ProcessError()

	def start (self, restart=False):
		for process in self.reactor.configuration.process:
			if restart and process in list(self._process):
				self._terminate(process)
			self._start(process)
		for process in list(self._process):
			if process not in self.reactor.configuration.process:
				self._terminate(process)

	def fds (self):
		return [self._process[process].stdout for process in self._process]

	def received (self):
		consumed_data = False

		for process in list(self._process):
			try:
				proc = self._process[process]
				# proc.poll returns None if the process is still fine
				# -[signal], like -15, if the process was terminated
				if proc.poll() is not None and self.reactor.respawn:
					raise ValueError('child died')
				r,_,_ = select.select([proc.stdout,],[],[],0)
				if r:
					try:
						while True:
							# Calling next() on Linux and OSX works perfectly well
							# but not on OpenBSD where it always raise StopIteration
							# and only readline() works
							raw = proc.stdout.readline()
							line = raw.rstrip()
							consumed_data = True
							self.logger.processes("Command from process %s : %s " % (process,line))
							if raw == '':
								raise IOError('Child process died')
							yield (process,formated(line))
					except IOError,exc:
						if not exc.errno or exc.errno in error.fatal:
							# if the program exists we can get an IOError with errno code zero !
							self.logger.processes("Issue with the process' PIPE, terminating it and restarting it")
							self._terminate(process)
							self._start(process)
						elif exc.errno in error.block:
							# we often see errno.EINTR: call interrupted and
							# we most likely have data, we will try to read them a the next loop iteration
							pass
						else:
							self.logger.processes("unexpected errno received from forked process (%s)" % errstr(exc))
					except StopIteration:
						if not consumed_data:
							self.logger.processes("The process died, trying to respawn it")
							self._terminate(process)
							self._start(process)
			except (subprocess.CalledProcessError,OSError,ValueError):
				self.logger.processes("Issue with the process, terminating it and restarting it")
				self._terminate(process)
				self._start(process)

	def write (self, process, string, peer=None):
		failure = 0
		while True:
			try:
				self._process[process].stdin.write('%s\n' % string)
				self._process[process].stdin.flush()
			except IOError,exc:
				failure += 1
				if failure >= 5:
					self.logger.processes("Too many attempt to send data to helper program, aborting")
					raise ProcessError()

				if exc.errno in error.block:
					# Could it have been caused by a signal ? What to do.
					self.logger.processes("Error received while sending data to helper program, retrying (%s)" % errstr(exc))
					continue

				if exc.errno == errno.EPIPE:
					self.logger.processes("Issue while sending data to our helper program, it left us restarting it")
					self._terminate(process)
					self._start(process)
					continue

				self.logger.processes("Fatal error on writing to PIPE, ignoring as it can not recovered")
				raise ProcessError()
			break
		return True

	def _notify (self, peer, event):
		neighbor = peer.neighbor.peer_address
		for process in self._neighbor_process.get(neighbor,[]):
			if process in self._process:
				if event in self._events[process]:
					yield process
		for process in self._neighbor_process.get('*',[]):
			if process in self._process:
				if event in self._events[process]:
					yield process

	# do not do anything if silenced
	# no-self-argument

	def silenced (function):
		def closure (self, *args):
			if self.silence:
				return
			return function(self,*args)
		return closure

	# invalid-name
	@silenced
	def up (self, peer):
		for process in self._notify(peer,'neighbor-changes'):
			self.write(process,self._encoder[process].up(peer),peer)

	@silenced
	def connected (self, peer):
		for process in self._notify(peer,'neighbor-changes'):
			self.write(process,self._encoder[process].connected(peer),peer)

	@silenced
	def down (self, peer, reason):
		for process in self._notify(peer,'neighbor-changes'):
			self.write(process,self._encoder[process].down(peer,reason),peer)

	@silenced
	def receive (self, peer, category, header, body):
		for process in self._notify(peer,'receive-packets'):
			self.write(process,self._encoder[process].receive(peer,category,header,body),peer)

	@silenced
	def send (self, peer, category, header, body):
		for process in self._notify(peer,'send-packets'):
			self.write(process,self._encoder[process].send(peer,category,header,body),peer)

	@silenced
	def notification (self, peer, code, subcode, data):
		for process in self._notify(peer,'neighbor-changes'):
			self.write(process,self._encoder[process].notification(peer,code,subcode,data),peer)

	@silenced
	def message (self, message_id, peer, message, header,*body):
		self._dispatch[message_id](self,peer,message,header,*body)

	# registering message functions
	# no-self-argument

	def register_process (message_id, storage):
		def closure (function):
			def wrap (*args):
				function(*args)
			storage[message_id] = wrap
			return wrap
		return closure

	@register_process(Message.CODE.OPEN,_dispatch)
	def _open (self, peer, open_msg, header, body, direction='received'):
		for process in self._notify(peer,'receive-opens'):
			self.write(process,self._encoder[process].open(peer,direction,open_msg,header,body),peer)

	# unused-argument, must keep the API
	@register_process(Message.CODE.KEEPALIVE,_dispatch)
	def _keepalive (self, peer, keepalive, header, body):
		for process in self._notify(peer,'receive-keepalives'):
			self.write(process,self._encoder[process].keepalive(peer,header,body),peer)

	@register_process(Message.CODE.UPDATE,_dispatch)
	def _update (self, peer, update, header, body):
		for process in self._notify(peer,'receive-updates'):
			self.write(process,self._encoder[process].update(peer,update,header,body),peer)

	@register_process(Message.CODE.ROUTE_REFRESH,_dispatch)
	def _refresh (self, peer, refresh, header, body):
		for process in self._notify(peer,'receive-refresh'):
			self.write(process,self._encoder[process].refresh(peer,refresh,header,body),peer)

	@register_process(Message.CODE.OPERATIONAL,_dispatch)
	def _operational (self, peer, operational, header, body):
		for process in self._notify(peer,'receive-operational'):
			self.write(process,self._encoder[process].operational(peer,operational.category,operational,header,body),peer)

	@register_process(Message.CODE.NOTIFICATION,_dispatch)
	def _notification (self, peer, message, header, body):
		for process in self._notify(peer,'receive-notifications'):
			self.write(process,self._encoder[process].notification(peer,message.code,message.subcode,body),peer)



# encoding: utf-8
"""
clusterlist.py

Created by Thomas Mangin on 2012-07-07.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.protocol.ip import IPv4

from exabgp.bgp.message.update.attribute.attribute import Attribute


# ===================================================================
#

class ClusterID (IPv4):
	def __init__ (self, ip):
		IPv4.__init__(self,ip)


class ClusterList (Attribute):
	ID = Attribute.CODE.CLUSTER_LIST
	FLAG = Attribute.Flag.OPTIONAL
	CACHING = True

	__slots__ = ['clusters','packed','_len']

	def __init__ (self, clusters, packed=None):
		self.clusters = clusters
		self.packed = self._attribute(packed if packed else ''.join([_.pack() for _ in clusters]))
		self._len = len(clusters)*4

	def pack (self, negotiated=None):
		return self.packed

	def __len__ (self):
		return self._len

	def __str__ (self):
		if self._len != 1:
			return '[ %s ]' % ' '.join([str(_) for _ in self.clusters])
		return '%s' % self.clusters[0]

	def json (self):
		return '[ %s ]' % ', '.join(['"%s"' % str(_) for _ in self.clusters])

	@classmethod
	def unpack (cls, data, negotiated):
		clusters = []
		while data:
			clusters.append(IPv4.unpack(data[:4]))
			data = data[4:]
		return cls(clusters)



# encoding: utf-8
"""
message.py

Created by Thomas Mangin on 2013-02-26.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""


class Message (int):
	ROUTE_MONITORING = 0
	STATISTICS_REPORT = 1
	PEER_DOWN_NOTIFICATION = 2

	_str = {
		0: 'route monitoring',
		1: 'statistics report',
		2: 'peer down notification',
	}

	def __str__ (self):
		return self._str.get(self,'unknow %d' % self)

	def validate (self):
		return self in (0,1,2)

stat = {
	0: "prefixes rejected by inbound policy",
	1: "(known) duplicate prefix advertisements",
	2: "(known) duplicate withdraws",
	3: "updates invalidated due to CLUSTER_LIST loop",
	4: "updates invalidated due to AS_PATH loop",
}

peer = {
	1: "Local system closed session, notification sent",
	2: "Local system closed session, no notification",
	3: "Remote system closed session, notification sent",
	4: "Remote system closed session, no notification",
}



# encoding: utf-8
"""
atomicaggregate.py

Created by Thomas Mangin on 2012-07-14.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.bgp.message.update.attribute.attribute import Attribute
from exabgp.bgp.message.notification import Notify


# ========================================================== AtomicAggregate (6)
#

class AtomicAggregate (Attribute):
	ID = Attribute.CODE.ATOMIC_AGGREGATE
	FLAG = Attribute.Flag.TRANSITIVE
	CACHING = True

	__slots__ = []

	def pack (self, negotiated=None):
		return self._attribute('')

	def __len__ (self):
		return 0

	def __str__ (self):
		return ''

	def __cmp__ (self, other):
		if not isinstance(other,self.__class__):
			return -1
		return 0

	def __hash__ (self):
		return 0

	@classmethod
	def unpack (cls, data, negotiated):
		if data:
			raise Notify(3,2,'invalid ATOMIC_AGGREGATE %s' % [hex(ord(_)) for _ in data])
		return cls()

	@classmethod
	def setCache (cls):
		# There can only be one, build it now :)
		cls.cache[Attribute.CODE.ATOMIC_AGGREGATE][''] = cls()

AtomicAggregate.setCache()



# encoding: utf-8
"""
neighbor.py

Created by Thomas Mangin on 2014-06-22.
Copyright (c) 2014-2015 Exa Networks. All rights reserved.
"""

from exabgp.configuration.engine.raised import Raised
from exabgp.configuration.engine.section import Section
from exabgp.configuration.engine.parser import ip
from exabgp.configuration.engine.parser import ttl
from exabgp.configuration.engine.parser import md5

from exabgp.configuration.bgp.session import SectionSession


# ============================================================== syntax_neighbor

syntax_neighbor = """\
neighbor <name> {
	session <name>
	tcp {
		bind          82.219.212.34
		connect       195.8.215.15"
		ttl-security  disable
		md5           "secret"
	}
	announce [
		local-routes
		off-goes-the-ddos
	]
}
"""

# =============================================================== RaisedNeighbor

class RaisedNeighbor (Raised):
	syntax = syntax_neighbor


# ============================================================== SectionNeighbor


class SectionNeighbor (Section):
	syntax = syntax_neighbor
	name = 'neighbor'

	def enter (self, tokeniser):
		Section.enter(self,tokeniser)

	def exit (self, tokeniser):
		if 'tcp-bind' not in self.content:
			raise RaisedNeighbor('neighbor needs a tcp bind ip')

		if 'tcp-connect' not in self.content:
			raise RaisedNeighbor('neighbor needs a tcp connect ip')

		if 'tcp-ttl-security' not in self.content:
			self.content['tcp-ttl-security'] = None

		if 'tcp-md5' not in self.content:
			self.content['tcp-md5'] = None

	def session (self, tokeniser):
		section = self.get_section(SectionSession.name,tokeniser)
		if section:
			self.content['session'] = section
		else:
			return False

	def announce (self, tokeniser):
		announced = tokeniser()
		if not hasattr(announced,'pop'):
			raise RaisedNeighbor('announce takes a list of named routes')
		self.content['announce'] = [word for (line,column,line,word) in announced]

	def tcp_bind (self, tokeniser):
		try:
			self.content['tcp-bind'] = ip(tokeniser)
		except ValueError,exc:
			raise RaisedNeighbor(tokeniser,'could not parse tcp bind ip, %s' % str(exc))

	def tcp_connect (self, tokeniser):
		try:
			self.content['tcp-connect'] = ip(tokeniser)
		except ValueError,exc:
			raise RaisedNeighbor(tokeniser,'could not parse tcp connect ip, %s' % str(exc))

	def tcp_ttl_security (self, tokeniser):
		try:
			self.content['tcp-ttl-security'] = ttl(tokeniser)
		except ValueError,exc:
			raise RaisedNeighbor(tokeniser,'could not parse tcp ttl, %s' % str(exc))

	def tcp_md5 (self, tokeniser):
		try:
			self.content['tcp-md5'] = md5(tokeniser)
		except ValueError,exc:
			raise RaisedNeighbor(tokeniser,'could not parse tcp MD5, %s' % str(exc))

	@classmethod
	def register (cls, registry, location):
		registry.register_class(cls)

		registry.register_hook(cls,'enter',location,'enter')

		registry.register(SectionSession,location+['session'])
		registry.register_hook(cls,'action',location+['session'],'session')

		registry.register_hook(cls,'action',location+['announce'],'announce')

		for tcp in (location+['tcp'],):
			registry.register_hook(cls,'enter',tcp,'enter_nameless')
			registry.register_hook(cls,'action',tcp+['bind'],'tcp_bind')
			registry.register_hook(cls,'action',tcp+['connect'],'tcp_connect')
			registry.register_hook(cls,'action',tcp+['ttl-security'],'tcp_ttl_security')
			registry.register_hook(cls,'action',tcp+['md5'],'tcp_md5')
			registry.register_hook(cls,'exit', tcp,'exit')


		registry.register_hook(cls,'exit',location,'exit')



# encoding: utf-8
"""
version.py

Created by Thomas Mangin on 2012-07-17.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

# =================================================================== Version
#


class Version (int):
	def pack (self):
		return chr(self)



# encoding: utf-8
"""
icmp.py

Created by Thomas Mangin on 2010-01-15.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.protocol.enum import Enum

# ============================================================== ICMP Code Field
# http://www.iana.org/assignments/icmp-parameters

class ICMPType (Enum):
	ECHO_REPLY               = 0x00
	DESTINATION_UNREACHEABLE = 0x03
	SOURCE_QUENCH            = 0x04
	REDIRECT                 = 0x05
	ECHO_REQUEST             = 0x08
	ROUTER_ADVERTISEMENT     = 0x09
	ROUTER_SOLICIT           = 0x0A
	TIME_EXCEEDED            = 0x0B
	PARAMETER_PROBLEM        = 0x0C
	TIMESTAMP                = 0x0D
	TIMESTAMP_REPLY          = 0x0E
	INFO_REQUEST             = 0x0F
	INFO_REPLY               = 0x10
	MASK_REQUEST             = 0x11
	MASK_REPLY               = 0x12
	TRACEROUTE               = 0x1E


ICMPType.UNKNOWN = 'unknown icmp type %s'

ICMPType.NAME = {
	ICMPType(ICMPType.ECHO_REPLY):               'echo-reply',
	ICMPType(ICMPType.ECHO_REQUEST):             'echo-request',
	ICMPType(ICMPType.INFO_REPLY):               'info-reply',
	ICMPType(ICMPType.INFO_REQUEST):             'info-request',
	ICMPType(ICMPType.MASK_REPLY):               'mask-reply',
	ICMPType(ICMPType.MASK_REQUEST):             'mask-request',
	ICMPType(ICMPType.PARAMETER_PROBLEM):        'parameter-problem',
	ICMPType(ICMPType.REDIRECT):                 'redirect',
	ICMPType(ICMPType.ROUTER_ADVERTISEMENT):     'router-advertisement',
	ICMPType(ICMPType.ROUTER_SOLICIT):           'router-solicit',
	ICMPType(ICMPType.SOURCE_QUENCH):            'source-quench',
	ICMPType(ICMPType.TIME_EXCEEDED):            'time-exceeded',
	ICMPType(ICMPType.TIMESTAMP):                'timestamp',
	ICMPType(ICMPType.TIMESTAMP_REPLY):          'timestamp-reply',
	ICMPType(ICMPType.DESTINATION_UNREACHEABLE): 'unreachable',
}

ICMPType.VALUE = {
	'echo-reply':               ICMPType(ICMPType.ECHO_REPLY),
	'echo-request':             ICMPType(ICMPType.ECHO_REQUEST),
	'info-reply':               ICMPType(ICMPType.INFO_REPLY),
	'info-request':             ICMPType(ICMPType.INFO_REQUEST),
	'mask-reply':               ICMPType(ICMPType.MASK_REPLY),
	'mask-request':             ICMPType(ICMPType.MASK_REQUEST),
	'parameter-problem':        ICMPType(ICMPType.PARAMETER_PROBLEM),
	'redirect':                 ICMPType(ICMPType.REDIRECT),
	'router-advertisement':     ICMPType(ICMPType.ROUTER_ADVERTISEMENT),
	'router-solicit':           ICMPType(ICMPType.ROUTER_SOLICIT),
	'source-quench':            ICMPType(ICMPType.SOURCE_QUENCH),
	'time-exceeded':            ICMPType(ICMPType.TIME_EXCEEDED),
	'timestamp':                ICMPType(ICMPType.TIMESTAMP),
	'timestamp-reply':          ICMPType(ICMPType.TIMESTAMP_REPLY),
	'unreachable': ICMPType(ICMPType.DESTINATION_UNREACHEABLE),
}


# http://www.iana.org/assignments/icmp-parameters
class ICMPCode (Enum):
	# Destination Unreacheable (type 3)
	NETWORK_UNREACHABLE                   = 0x0
	HOST_UNREACHABLE                      = 0x1
	PROTOCOL_UNREACHABLE                  = 0x2
	PORT_UNREACHABLE                      = 0x3
	FRAGMENTATION_NEEDED                  = 0x4
	SOURCE_ROUTE_FAILED                   = 0x5
	DESTINATION_NETWORK_UNKNOWN           = 0x6
	DESTINATION_HOST_UNKNOWN              = 0x7
	SOURCE_HOST_ISOLATED                  = 0x8
	DESTINATION_NETWORK_PROHIBITED        = 0x9
	DESTINATION_HOST_PROHIBITED           = 0xA
	NETWORK_UNREACHABLE_FOR_TOS           = 0xB
	HOST_UNREACHABLE_FOR_TOS              = 0xC
	COMMUNICATION_PROHIBITED_BY_FILTERING = 0xD
	HOST_PRECEDENCE_VIOLATION             = 0xE
	PRECEDENCE_CUTOFF_IN_EFFECT           = 0xF

	# Redirect (Type 5)
	REDIRECT_FOR_NETWORK                  = 0x0
	REDIRECT_FOR_HOST                     = 0x1
	REDIRECT_FOR_TOS_AND_NET              = 0x2
	REDIRECT_FOR_TOS_AND_HOST             = 0x3

	# Time Exceeded (Type 11)
	TTL_EQ_ZERO_DURING_TRANSIT            = 0x0
	TTL_EQ_ZERO_DURING_REASSEMBLY         = 0x1

	# parameter Problem (Type 12)
	REQUIRED_OPTION_MISSING               = 0x1
	IP_HEADER_BAD                         = 0x2


ICMPCode.UNKNOWN = 'unknown icmp code %s'

ICMPCode.VALUE = {
	'communication-prohibited-by-filtering': ICMPCode(ICMPCode.COMMUNICATION_PROHIBITED_BY_FILTERING),
	'destination-host-prohibited':           ICMPCode(ICMPCode.DESTINATION_HOST_PROHIBITED),
	'destination-host-unknown':              ICMPCode(ICMPCode.DESTINATION_HOST_UNKNOWN),
	'destination-network-prohibited':        ICMPCode(ICMPCode.DESTINATION_NETWORK_PROHIBITED),
	'destination-network-unknown':           ICMPCode(ICMPCode.DESTINATION_NETWORK_UNKNOWN),
	'fragmentation-needed':                  ICMPCode(ICMPCode.FRAGMENTATION_NEEDED),
	'host-precedence-violation':             ICMPCode(ICMPCode.HOST_PRECEDENCE_VIOLATION),
	'host-unreachable':                      ICMPCode(ICMPCode.HOST_UNREACHABLE),
	'host-unreachable-for-tos':              ICMPCode(ICMPCode.HOST_UNREACHABLE_FOR_TOS),
	'ip-header-bad':                         ICMPCode(ICMPCode.IP_HEADER_BAD),
	'network-unreachable':                   ICMPCode(ICMPCode.NETWORK_UNREACHABLE),
	'network-unreachable-for-tos':           ICMPCode(ICMPCode.NETWORK_UNREACHABLE_FOR_TOS),
	'port-unreachable':                      ICMPCode(ICMPCode.PORT_UNREACHABLE),
	'precedence-cutoff-in-effect':           ICMPCode(ICMPCode.PRECEDENCE_CUTOFF_IN_EFFECT),
	'protocol-unreachable':                  ICMPCode(ICMPCode.PROTOCOL_UNREACHABLE),
	'redirect-for-host':                     ICMPCode(ICMPCode.REDIRECT_FOR_HOST),
	'redirect-for-network':                  ICMPCode(ICMPCode.REDIRECT_FOR_NETWORK),
	'redirect-for-tos-and-host':             ICMPCode(ICMPCode.REDIRECT_FOR_TOS_AND_HOST),
	'redirect-for-tos-and-net':              ICMPCode(ICMPCode.REDIRECT_FOR_TOS_AND_NET),
	'required-option-missing':               ICMPCode(ICMPCode.REQUIRED_OPTION_MISSING),
	'source-host-isolated':                  ICMPCode(ICMPCode.SOURCE_HOST_ISOLATED),
	'source-route-failed':                   ICMPCode(ICMPCode.SOURCE_ROUTE_FAILED),
	'ttl-eq-zero-during-reassembly':         ICMPCode(ICMPCode.TTL_EQ_ZERO_DURING_REASSEMBLY),
	'ttl-eq-zero-during-transit':            ICMPCode(ICMPCode.TTL_EQ_ZERO_DURING_TRANSIT),
}

ICMPCode.NAME = {
	ICMPCode(ICMPCode.COMMUNICATION_PROHIBITED_BY_FILTERING): 'communication-prohibited-by-filtering',
	ICMPCode(ICMPCode.DESTINATION_HOST_PROHIBITED):           'destination-host-prohibited',
	ICMPCode(ICMPCode.DESTINATION_HOST_UNKNOWN):              'destination-host-unknown',
	ICMPCode(ICMPCode.DESTINATION_NETWORK_PROHIBITED):        'destination-network-prohibited',
	ICMPCode(ICMPCode.DESTINATION_NETWORK_UNKNOWN):           'destination-network-unknown',
	ICMPCode(ICMPCode.FRAGMENTATION_NEEDED):                  'fragmentation-needed',
	ICMPCode(ICMPCode.HOST_PRECEDENCE_VIOLATION):             'host-precedence-violation',
	ICMPCode(ICMPCode.HOST_UNREACHABLE):                      'host-unreachable',
	ICMPCode(ICMPCode.HOST_UNREACHABLE_FOR_TOS):              'host-unreachable-for-tos',
	ICMPCode(ICMPCode.IP_HEADER_BAD):                         'ip-header-bad',
	ICMPCode(ICMPCode.NETWORK_UNREACHABLE):                   'network-unreachable',
	ICMPCode(ICMPCode.NETWORK_UNREACHABLE_FOR_TOS):           'network-unreachable-for-tos',
	ICMPCode(ICMPCode.PORT_UNREACHABLE):                      'port-unreachable',
	ICMPCode(ICMPCode.PRECEDENCE_CUTOFF_IN_EFFECT):           'precedence-cutoff-in-effect',
	ICMPCode(ICMPCode.PROTOCOL_UNREACHABLE):                  'protocol-unreachable',
	ICMPCode(ICMPCode.REDIRECT_FOR_HOST):                     'redirect-for-host',
	ICMPCode(ICMPCode.REDIRECT_FOR_NETWORK):                  'redirect-for-network',
	ICMPCode(ICMPCode.REDIRECT_FOR_TOS_AND_HOST):             'redirect-for-tos-and-host',
	ICMPCode(ICMPCode.REDIRECT_FOR_TOS_AND_NET):              'redirect-for-tos-and-net',
	ICMPCode(ICMPCode.REQUIRED_OPTION_MISSING):               'required-option-missing',
	ICMPCode(ICMPCode.SOURCE_HOST_ISOLATED):                  'source-host-isolated',
	ICMPCode(ICMPCode.SOURCE_ROUTE_FAILED):                   'source-route-failed',
	ICMPCode(ICMPCode.TTL_EQ_ZERO_DURING_REASSEMBLY):         'ttl-eq-zero-during-reassembly',
	ICMPCode(ICMPCode.TTL_EQ_ZERO_DURING_TRANSIT):            'ttl-eq-zero-during-transit',
}
